import os
import re
import uuid
import warnings
from collections import defaultdict
from datetime import datetime, date
import pandas as pd
from typing import Optional, Tuple, Union

from onetick.py.otq import otq

import onetick.py.functions
import onetick.py.sources
from onetick import py as otp
from onetick.py import types as ott
from onetick.py import utils, configuration
from onetick.py.core._internal._manually_bound_value import _ManuallyBoundValue
from onetick.py.core._internal._proxy_node import _ProxyNode
from onetick.py.core._internal._state_vars import StateVars
from onetick.py.core._source._symbol_param import _SymbolParamColumn, _SymbolParamSource
from onetick.py.core._source.schema import Schema
from onetick.py.core._source.symbol import Symbol
from onetick.py.core._source.tmp_otq import TmpOtq
from onetick.py.core.column import _Column
from onetick.py.core.column_operations.base import _Operation, OnetickParameter
from onetick.py.core.query_inspector import get_query_parameter_list
from onetick.py.utils import adaptive, adaptive_to_default, default, render_otq


def _is_dict_required(symbols):
    """
    Depending on symbols, determine if output of otp.run() or Source.__call__() should always be a dictionary
    of {symbol: dataframe} even if only one symbol is present in the results
    """
    if isinstance(symbols, (list, tuple)):
        if len(symbols) == 0:
            return False
        elif len(symbols) > 1:
            return True
        else:
            symbols = symbols[0]

    if isinstance(symbols, otp.Source):
        return True
    if isinstance(symbols, otq.Symbol):
        symbols = symbols.name
    if isinstance(symbols, str) and 'eval' in symbols:
        return True
    return False


class MetaFields:
    """
    OneTick defines several pseudo-columns that can be treated as if they were columns of every tick.

    These columns can be accessed directly via :py:meth:`onetick.py.Source.__getitem__` method.

    But in case they are used in :py:class:`~onetick.py.core.column_operations.base.Expr`
    they can be accessed via ``onetick.py.Source.meta_fields``.

    Examples
    --------

    Accessing pseudo-fields as columns or as class properties

    >>> data = otp.Tick(A=1)
    >>> data['X'] = data['_START_TIME']
    >>> data['Y'] = otp.Source.meta_fields['_TIMEZONE']
    >>> otp.run(data, start=otp.dt(2003, 12, 2), timezone='GMT')
            Time  A          X    Y
    0 2003-12-02  1 2003-12-02  GMT
    """
    def __init__(self):
        self.timestamp = _Column('TIMESTAMP', dtype=ott.nsectime)
        self.time = self.timestamp
        self.start_time = _Column('_START_TIME', dtype=ott.nsectime)
        self.start = self.start_time
        self.end_time = _Column('_END_TIME', dtype=ott.nsectime)
        self.end = self.end_time
        self.timezone = _Column('_TIMEZONE', dtype=str)
        self.db_name = _Column('_DBNAME', dtype=str)
        self.symbol_name = _Column('_SYMBOL_NAME', dtype=str)
        self.tick_type = _Column('_TICK_TYPE', dtype=str)
        self.symbol_time = _Column('_SYMBOL_TIME', dtype=otp.nsectime)
        self.__fields = set(map(str, self.__dict__.values())) | {'Time'}

    def get_onetick_fields_and_types(self):
        return {
            column.name: column.dtype
            for name, column in self.__dict__.items()
            if not name.startswith('_') and name != 'time'
        }

    def __iter__(self):
        yield from self.__fields

    def __contains__(self, item):
        return item in self.__fields

    def __len__(self):
        return len(self.__fields)

    def __getitem__(self, item):
        """
        These fields are available:

        * ``TIMESTAMP`` (or ``Time``)
        * ``START_TIME`` (or ``_START_TIME``)
        * ``END_TIME`` (or ``_END_TIME``)
        * ``TIMEZONE`` (or ``_TIMEZONE``)
        * ``DBNAME`` (or ``_DBNAME``)
        * ``SYMBOL_NAME`` (or ``_SYMBOL_NAME``)
        * ``TICK_TYPE`` (or ``_TICK_TYPE``)
        * ``SYMBOL_TIME`` (or ``_SYMBOL_TIME``)
        """
        return {
            'TIMESTAMP': self.timestamp,
            'Time': self.time,
            'START_TIME': self.start_time,
            '_START_TIME': self.start_time,
            'END_TIME': self.end_time,
            '_END_TIME': self.end_time,
            'TIMEZONE': self.timezone,
            '_TIMEZONE': self.timezone,
            'DB_NAME': self.db_name,
            'DBNAME': self.db_name,
            '_DBNAME': self.db_name,
            'SYMBOL_NAME': self.symbol_name,
            '_SYMBOL_NAME': self.symbol_name,
            'TICK_TYPE': self.tick_type,
            '_TICK_TYPE': self.tick_type,
            'SYMBOL_TIME': self.symbol_time,
            '_SYMBOL_TIME': self.symbol_time,
        }[item]


class Source:
    """
    Base class for representing Onetick execution graph.
    All :ref:`onetick-py sources <api/sources/root:sources>` are derived from this class
    and have access to all its methods.

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> isinstance(data, otp.Source)
    True

    Also this class can be used to initialize raw source
    with the help of ``onetick.query`` classes, but
    it should be done with caution as the user is required to set
    such properties as symbol name and tick type manually.

    >>> data = otp.Source(otq.TickGenerator(bucket_interval=0, fields='long A = 123').tick_type('TT'))
    >>> otp.run(data, symbols='LOCAL::')
            Time    A
    0 2003-12-04  123
    """

    # TODO: need to support transactions for every _source
    # transaction is set of calls between _source creation and call or between two calls
    # if transaction have the same operations, then it seems we should add only one set of operations

    _PROPERTIES = [
        "__node",
        "__hash",
        "__sources_keys_dates",
        "__sources_modify_query_times",
        "__sources_base_ep_func",
        "__sources_symbols",
        "__source_has_output",
        "__name",
        "_tmp_otq"
    ]
    _OT_META_FIELDS = ["_START_TIME", "_END_TIME", "_SYMBOL_NAME", "_DBNAME", "_TICK_TYPE", '_TIMEZONE']
    meta_fields = MetaFields()
    Symbol = Symbol  # NOSONAR

    def __init__(
        self,
        node=None,
        schema=None,
        _symbols=None,
        _start=adaptive,
        _end=adaptive,
        _base_ep_func=None,
        _has_output=True,
        **kwargs,
    ):

        self._tmp_otq = TmpOtq()
        self.__name = None

        if isinstance(_symbols, OnetickParameter):
            _symbols = _symbols.parameter_expression

        schema = self._select_schema(schema, kwargs)

        for key in schema:
            if self._check_key_in_properties(key):
                raise ValueError(f"Can't set class property {key}")
            if self._check_key_is_meta(key):
                if key == 'TIMESTAMP':
                    # for backward-compatibility
                    warnings.warn(f"Setting meta field {key} in schema is not needed", FutureWarning, stacklevel=2)
                else:
                    raise ValueError(f"Can't set meta field {key}")

        schema.update(
            self.meta_fields.get_onetick_fields_and_types()
        )

        for key, value in schema.items():
            # calculate value type
            value_type = ott.get_source_base_type(value)
            self.__dict__[key] = _Column(name=key, dtype=value_type, obj_ref=self)

        # just an alias to Timestamp
        self.__dict__['Time'] = self.__dict__['TIMESTAMP']
        self.__dict__['_state_vars'] = StateVars(self)

        if node is None:
            node = otq.Passthrough()

        if isinstance(_symbols, _SymbolParamColumn):
            symbol_node = otq.ModifySymbolName(str(self.Symbol[_symbols.name, str]))
            node = node.sink(symbol_node)
            _symbols = None

        self.__hash = uuid.uuid4()
        self.__sources_keys_dates = {}
        self.__sources_modify_query_times = {}
        self.__sources_base_ep_func = {}
        self.__sources_symbols = {}
        self.__source_has_output = _has_output

        if isinstance(node, _ProxyNode):
            self.__node = _ProxyNode(*node.copy_graph(), refresh_func=self.__refresh_hash)
        else:
            self.__node = _ProxyNode(*self.__from_ep_to_proxy(node), refresh_func=self.__refresh_hash)
            self.__sources_keys_dates[self.__node.key()] = (_start, _end)
            self.__sources_modify_query_times[self.__node.key()] = False
            self.__sources_base_ep_func[self.__node.key()] = _base_ep_func
            self.__sources_symbols[self.__node.key()] = _symbols

    def _try_default_constructor(self, *args, node=None, schema=None, **kwargs):
        if node is not None:
            # Source.copy() method will use this way
            # all info from original source will be copied by copy() method after
            Source.__init__(self, *args, node=node, schema=schema, **kwargs)
            return True
        return False

    def base_ep(self, **_kwargs):
        # default implementation
        # real implementation should return a Source object
        return None

    def _select_schema(self, schema, kwargs) -> dict:
        """
        Selects schema definition from ``schema`` or ``kwargs`` parameters
        for the time of deprecation of ``kwargs`` parameter.
        """
        if schema is None:
            if kwargs:
                warnings.warn(
                    f'Setting `{self.__class__.__name__}` schema via `**kwargs` is deprecated. '
                    'Please use `schema` parameter for this. '
                    f'Passed kwargs are: {kwargs}.',
                    FutureWarning,
                    stacklevel=2,
                )
                return kwargs
            else:
                return {}
        elif kwargs:
            raise ValueError(
                "Specifying schema through both `**kwargs` and `schema` is prohibited. "
                f"Passed kwargs are: {kwargs}."
            )

        return schema.copy()

    def _clean_sources_dates(self):
        self.__sources_keys_dates = {}
        self.__sources_modify_query_times = {}
        self.__sources_base_ep_func = {}
        self.__sources_symbols = {}

    def _set_sources_dates(self, other, copy_symbols=True):
        self.__sources_keys_dates.update(other._get_sources_dates())
        self.__sources_modify_query_times.update(other._get_sources_modify_query_times())
        self.__sources_base_ep_func.update(other._get_sources_base_ep_func())
        if copy_symbols:
            self.__sources_symbols.update(other._get_sources_symbols())
        else:
            # this branch is applicable for the bound symbols with callbacks,
            # where we drop all adaptive symbols and keep only manually specified
            # symbols
            manually_bound = {
                key: _ManuallyBoundValue(value)
                for key, value in other._get_sources_symbols().items()
                if value is not adaptive and value is not adaptive_to_default
            }
            self.__sources_symbols.update(manually_bound)

        self.__source_has_output = other._get_source_has_output()

    def _change_sources_keys(self, keys: dict):
        """
        Change keys in sources dictionaries.
        Need to do it, for example, after rebuilding the node history with new keys.

        Parameters
        ----------
        keys: dict
            Mapping from old key to new key
        """
        sources = (self.__sources_keys_dates,
                   self.__sources_modify_query_times,
                   self.__sources_base_ep_func,
                   self.__sources_symbols)
        for dictionary in sources:
            for key in list(dictionary):
                dictionary[keys[key]] = dictionary.pop(key)

    def _get_source_has_output(self):
        return self.__source_has_output

    def _get_sources_dates(self):
        return self.__sources_keys_dates

    def _get_sources_modify_query_times(self):
        return self.__sources_modify_query_times

    def _get_sources_base_ep_func(self):
        return self.__sources_base_ep_func

    def _get_sources_symbols(self):
        return self.__sources_symbols

    def _check_key_in_properties(self, key: str) -> bool:
        if key in self.__class__._PROPERTIES:
            return True
        if key.replace('_' + Source.__name__.lstrip('_'), "") in self.__class__._PROPERTIES:
            return True
        if key.replace(self.__class__.__name__, "") in self.__class__._PROPERTIES:
            return True
        return False

    def _check_key_is_meta(self, key: str) -> bool:
        return key in self.__class__.meta_fields

    def _check_key_is_reserved(self, key: str) -> bool:
        return self._check_key_in_properties(key) or self._check_key_is_meta(key)

    def _set_field_by_tuple(self, name, dtype):
        warnings.warn('Using _set_field_by_tuple() is not recommended,'
                      ' change your code to use otp.Source.schema object.', DeprecationWarning)
        if name not in self.__dict__:
            if dtype is None:
                raise KeyError(f'Column name {name} is not in the schema. Please, check that this column '
                               'is in the schema or add it using the .schema property')

            if name in (0, 1):
                raise ValueError(f"constant {name} are not supported for indexing for now, please use otp.Empty")
            if isinstance(name, (int, float)):
                raise ValueError("integer indexes are not supported")
            self.__dict__[name] = _Column(name, dtype, self)
        else:
            if not isinstance(self.__dict__[name], _Column):
                raise AttributeError(f"There is no '{name}' column")

            if dtype:
                type1, type2 = self.__dict__[name].dtype, dtype
                b_type1, b_type2 = ott.get_base_type(type1), ott.get_base_type(type2)

                if b_type1 != b_type2:
                    if {type1, type2} == {int, float}:
                        self.__dict__[name]._dtype = float
                    else:
                        raise Warning(
                            f"Column '{name}' was declared as '{type1}', but you want to change it to '{type2}', "
                            "that is not possible without setting type directly via assigning value"
                        )

                else:
                    if issubclass(b_type1, str):
                        t1_length = ott.string.DEFAULT_LENGTH if type1 is str else type1.length
                        t2_length = ott.string.DEFAULT_LENGTH if type2 is str else type2.length

                        self.__dict__[name]._dtype = type2 if t1_length < t2_length else type1
                    if {type1, type2} == {ott.nsectime, ott.msectime}:
                        self.__dict__[name]._dtype = ott.nsectime
        return self.__dict__[name]

    def _hash(self):
        return self.__hash

    def _merge_tmp_otq(self, source):
        self._tmp_otq.merge(source._tmp_otq)

    def __prepare_graph(self, symbols=None, start=None, end=None, has_output=False):
        # We copy object here, because we will change it according to passed
        # symbols and date ranges. For example, we can add modify_query_times EP
        # if it is necessary

        obj = self.copy()
        if has_output:
            obj.sink(otq.Passthrough())
        start, end, symbols = obj._set_date_range_and_symbols(symbols, start, end)
        if start is adaptive:
            start = None
        if end is adaptive:
            end = None
        if symbols is not None and isinstance(symbols, pd.DataFrame):
            symbols = utils.get_symbol_list_from_df(symbols)
        if symbols is not None and not isinstance(symbols, list):
            symbols = [symbols]
        elif symbols is None:
            symbols = []
        _symbols = []
        for sym in symbols:
            _symbols.append(self._convert_symbol_to_string(sym, tmp_otq=obj._tmp_otq, start=start, end=end))

        return obj, start, end, _symbols

    def to_otq(self, file_name=None, file_suffix=None, query_name=None, symbols=None, start=None, end=None,
               timezone=None, raw=None, add_passthrough=True,
               running=False,
               start_time_expression=None,
               end_time_expression=None,
               symbol_date=None,
               concurrency=None,
               batch_size=None):
        """
        Save :class:`otp.Source <onetick.py.Source>` object to .otq file and return path to the saved file.

        Parameters
        ----------
        file_name: str
            Absolute or relative path to the saved file.
            If ``None``, create temporary file and name it randomly.
        file_suffix: str
            Suffix to add to the saved file name (including extension).
            Can be specified if ``file_name`` is ``None``
            to distinguish between different temporary files.
            Default: ".to_otq.otq"
        query_name: str
            Name of the main query in the created file.
            If ``None``, take name from this Source object.
            If that name is empty, set name to "query".
        symbols: str, list, :pandas:`DataFrame <pandas.DataFrame>`, :class:`Source`
            symbols to save query with
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            start time to save query with
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            end time to save query with
        timezone: str
            timezone to save query with
        raw

            .. deprecated:: 1.4.17

        add_passthrough: bool
            will add :py:class:`onetick.query.Passthrough` event processor at the end of resulting graph
        running: bool
            Indicates whether a query is CEP or not.
        start_time_expression: str, :py:class:`~onetick.py.Operation`, optional
            Start time onetick expression of the query. If specified, it will take precedence over ``start``.
        end_time_expression: str, :py:class:`~onetick.py.Operation`, optional
            End time onetick expression of the query. If specified, it will take precedence over ``end``.
        symbol_date: :py:class:`otp.datetime <onetick.py.datetime>` or :py:class:`datetime.datetime` or int
            Symbol date for the query or integer in the YYYYMMDD format.
            Will be applied only to the main query.
        concurrency: int
            Concurrency set for the query.
            Will be applied only to the main query.
        batch_size: int
            Batch size set for the query.
            Will be applied only to the main query.

        Returns
        -------
        result: str
            Relative (if ``file_name`` is relative) or absolute path to the created query
            in the format ``file_name::query_name``

        Examples
        --------
        Create the .otq file from a :class:`otp.Source <onetick.py.Source>` object:

        >>> t = otp.Tick(A=1)
        >>> t.to_otq()  # doctest: +SKIP
        '/tmp/test_user/run_20251202_181018_11054/impetuous-bullfrog.to_otq.otq::query'
        """
        if raw is not None:
            warnings.warn('The "raw" flag is deprecated and makes no effect', FutureWarning)

        if timezone is None:
            timezone = configuration.config.tz

        file_path = str(file_name) if file_name is not None else None
        if file_suffix is None:
            file_suffix = self._name_suffix('to_otq.otq')

        if query_name is None:
            query_name = self.get_name(remove_invalid_symbols=True)
        if query_name is None:
            query_name = 'query'

        if isinstance(start, _Operation) and start_time_expression is None:
            start_time_expression = str(start)
            start = utils.adaptive
        if isinstance(end, _Operation) and end_time_expression is None:
            end_time_expression = str(end)
            end = utils.adaptive

        if isinstance(start_time_expression, _Operation):
            start_time_expression = str(start_time_expression)
        if isinstance(end_time_expression, _Operation):
            end_time_expression = str(end_time_expression)

        obj, start, end, symbols = self.__prepare_graph(symbols, start, end)

        graph = obj._to_graph(add_passthrough=add_passthrough)
        graph.set_symbols(symbols)

        return obj._tmp_otq.save_to_file(query=graph, query_name=query_name, file_path=file_path,
                                         file_suffix=file_suffix, start=start, end=end, timezone=timezone,
                                         running_query_flag=running,
                                         start_time_expression=start_time_expression,
                                         end_time_expression=end_time_expression,
                                         symbol_date=symbol_date,
                                         concurrency=concurrency, batch_size=batch_size)

    def _store_in_tmp_otq(self, tmp_otq, operation_suffix="tmp_query", symbols=None, start=None, end=None,
                          raw=None, add_passthrough=True, name=None, timezone=None, symbol_date=None,
                          concurrency=None, batch_size=None):
        """
        Adds this source to the tmp_otq storage

        Parameters
        ----------
        tmp_otq: TmpOtq
            Storage object
        operation_suffix: str
            Suffix string to be added to the autogenerated graph name in the otq file
        name: str, optional
            If specified, this ``name`` will be used to save query
            and ``suffix`` parameter will be ignored.

        Returns
        -------
        result: str
            String with the name of the saved graph (starting with THIS::)
        """
        if raw is not None:
            warnings.warn('The "raw" flag is deprecated and makes no effect', FutureWarning)

        obj, start, end, symbols = self.__prepare_graph(symbols, start, end)
        tmp_otq.merge(obj._tmp_otq)

        if isinstance(start, ott.dt):  # OT save_to_file checks for the datetime time
            start = datetime.fromtimestamp(start.timestamp())
        elif isinstance(start, date):
            start = datetime(start.year, start.month, start.day)
        if isinstance(end, ott.dt):
            end = datetime.fromtimestamp(end.timestamp())
        elif isinstance(end, date):
            end = datetime(end.year, end.month, end.day)

        if timezone is None:
            timezone = configuration.config.tz

        graph = obj._to_graph(add_passthrough=add_passthrough)
        graph.set_start_time(start)
        graph.set_end_time(end)
        graph.set_symbols(symbols)
        if timezone is not None:
            if otq.webapi:
                graph.set_timezone(timezone)
            else:
                graph.time_interval_properties().set_timezone(timezone)

        params = {}
        if symbol_date is not None:
            params['symbol_date'] = symbol_date
        if concurrency is not None:
            params['concurrency'] = concurrency
        if batch_size is not None:
            params['batch_size'] = batch_size
        suffix = self._name_suffix(suffix=operation_suffix, separator='__', remove_invalid_symbols=True)
        return tmp_otq.add_query(graph, suffix=suffix, name=name, params=params)

    def __refresh_hash(self):
        """
        This internal function refreshes hash for every graph modification.
        It is used only in _ProxyNode, because it tracks nodes changes
        """
        self.__hash = uuid.uuid3(uuid.NAMESPACE_DNS, str(self.__hash))

    def _prepare_for_execution(self, symbols=None, start=None, end=None, start_time_expression=None,
                               end_time_expression=None, timezone=None, has_output=None,
                               running_query_flag=None, require_dict=False, node_name=None,
                               symbol_date=None):
        if has_output is None:
            has_output = self.__source_has_output

        if timezone is None:
            timezone = configuration.config.tz

        obj, start, end, symbols = self.__prepare_graph(symbols, start, end, has_output)
        require_dict = require_dict or _is_dict_required(symbols)

        if node_name is None:
            node_name = 'SOURCE_CALL_MAIN_OUT_NODE'
            obj.node().node_name(node_name)

        graph = obj._to_graph(add_passthrough=False)

        graph.set_symbols(symbols)

        # create name and suffix for generated .otq file
        if otp.config.main_query_generated_filename:
            name = otp.config.main_query_generated_filename
            if name.endswith('.otq'):
                suffix = ''
            else:
                suffix = '.otq'
            force = True
        else:
            name = ''
            suffix = self._name_suffix('run.otq')
            force = False

        clean_up = default
        if otp.config.otq_debug_mode:
            clean_up = False
        base_dir = None
        if os.getenv('OTP_WEBAPI_TEST_MODE'):
            from onetick.py.otq import _tmp_otq_path
            base_dir = _tmp_otq_path()
        tmp_file = utils.TmpFile(name=name,
                                 suffix=suffix,
                                 force=force,
                                 base_dir=base_dir,
                                 clean_up=clean_up)

        query_to_run = obj._tmp_otq.save_to_file(query=graph,
                                                 query_name=self.get_name(remove_invalid_symbols=True)
                                                 if self.get_name(remove_invalid_symbols=True) else "main_query",
                                                 file_path=tmp_file.path,
                                                 start=start, end=end,
                                                 running_query_flag=running_query_flag,
                                                 start_time_expression=start_time_expression,
                                                 end_time_expression=end_time_expression,
                                                 timezone=timezone,
                                                 symbol_date=symbol_date)

        # PY-1423: we should set symbol_date in otp.run always
        symbol_date_to_run = None
        if symbol_date is not None:
            symbol_date_to_run = utils.symbol_date_to_str(symbol_date)
        else:
            symbol_dates_in_tmp_otq = obj._tmp_otq._get_symbol_dates()
            if symbol_dates_in_tmp_otq:
                symbol_date_to_run = symbol_dates_in_tmp_otq[0]
                if len(set(symbol_dates_in_tmp_otq)) > 1:
                    warnings.warn(
                        f'There are different symbol dates in resulting .otq file: {set(symbol_dates_in_tmp_otq)}.'
                        'But no symbol date were specified in otp.run.\n'
                        f'In this case the first symbol_date ({symbol_date_to_run}) will be used automatically.'
                    )

        return query_to_run, require_dict, node_name, symbol_date_to_run

    def __call__(self, *args, **kwargs):
        """
        .. deprecated:: 1.48.3
           Use :py:func:`otp.run <onetick.py.run>` instead.
        """
        warnings.warn('__call__() method is deprecated, use otp.run() instead', FutureWarning, stacklevel=2)
        return otp.run(self, *args, **kwargs)

    def to_df(self, symbols=None, **kwargs):
        """
        .. deprecated:: 1.48.3
           Use :py:func:`otp.run <onetick.py.run>` instead.
        """
        warnings.warn('to_df() method is deprecated, use otp.run() instead', FutureWarning, stacklevel=2)
        # For backward compatibility: otp.run() does not accept "symbols" as a non-keyword argument
        if symbols is not None:
            kwargs['symbols'] = symbols
        return otp.run(self, **kwargs)

    to_dataframe = to_df

    def print_api_graph(self):
        self.node().copy_graph(print_out=True)

    def _add_table(self, strict=False):
        table = otq.Table(
            fields=",".join(
                ott.type2str(dtype) + " " + name for name, dtype in self.columns(skip_meta_fields=True).items()
            ),
            keep_input_fields=not strict,
        )
        self.sink(table)

    def _is_unbound_required(self):
        """ Check whether a graph needs unbound symbol or not """

        for symbol in self.__sources_symbols.values():
            if symbol is adaptive or symbol is adaptive_to_default:
                return True
        return False

    def _get_widest_time_range(self):
        """
        Get minimum start time and maximum end time.
        If time is not found, None is returned.
        """
        start_times = []
        end_times = []

        for start, end in self.__sources_keys_dates.values():
            if start is not adaptive:
                start_times.append(start)
            if end is not adaptive:
                end_times.append(end)

        start = min(start_times) if start_times else None
        end = max(end_times) if end_times else None
        return start, end

    def __get_common_symbol(self):
        need_to_bind_symbol = False
        common_symbol = None

        # let's try to understand whether we could use common symbol for all sources
        # or we need to bound symbols instead
        first_symbol = None

        for symbol in self.__sources_symbols.values():
            if first_symbol is None:
                first_symbol = symbol

                if isinstance(first_symbol, _ManuallyBoundValue):
                    # Mark that we need to bound, but keep common_symbol equal to None.
                    # It is applicable for the bound symbols inside the merge with bound
                    # symbols, for example.
                    need_to_bind_symbol = True
                else:
                    common_symbol = symbol

                continue

            if symbol and symbol != first_symbol:
                need_to_bind_symbol = True
                common_symbol = None
                break

        # symbol is specified nowhere - just set unbound to the default one
        if (first_symbol is adaptive or first_symbol is adaptive_to_default) and (
            common_symbol is adaptive or common_symbol is adaptive_to_default
        ):
            common_symbol = configuration.config.default_symbol

        return common_symbol, need_to_bind_symbol

    def __get_modify_query_times(self, key, start, end, sources_start, sources_end):
        # determine whether we have to add modify_query_times to a src

        if self.__sources_modify_query_times[key]:
            return None

        start_date, end_date = self.__sources_keys_dates[key]

        if start_date is adaptive and end_date is adaptive:
            return None

        # if some of the end is specified, then it means
        # we need to check whether it is worth to wrap into the modify_query_times
        if start_date is adaptive:
            if start is None:
                start_date = sources_start
            else:
                start_date = start

        if end_date is adaptive:
            if end is None:
                end_date = sources_end
            else:
                end_date = end

        if start_date is adaptive or end_date is adaptive:
            return None

        # it might happen when either sources_start/end are adaptive
        # or start/end are adaptive
        if (
            (start is None and sources_start is not adaptive and start_date != sources_start)
            or (start is not None and start_date != start)
            or (end is None and sources_end is not adaptive and end_date != sources_end)
            or (end is not None and end_date != end)
        ):
            mqt_format = 'parse_time("%Y-%m-%d %H:%M:%S.%q","{}", _TIMEZONE)'
            mqt_ep = otq.ModifyQueryTimes(
                start_time=mqt_format.format(start_date.strftime("%Y-%m-%d %H:%M:%S.%f")),
                end_time=mqt_format.format(end_date.strftime("%Y-%m-%d %H:%M:%S.%f")),
                output_timestamp="TIMESTAMP",
            )

            return mqt_ep
        return None

    def _set_date_range_and_symbols(self, symbols=None, start=None, end=None):
        # will modify self

        if symbols is None:
            common_symbol, need_to_bind_symbol = self.__get_common_symbol()
        else:
            # when unbound symbols passed
            common_symbol = symbols
            need_to_bind_symbol = True  # use to check all sources whether some has bound symbols

        # Find max and min for _source data ranges
        sources_start, sources_end = self._get_widest_time_range()
        sources_start = sources_start or configuration.config.get('default_start_time', adaptive)
        sources_end = sources_end or configuration.config.get('default_end_time', adaptive)

        for key in self.__sources_keys_dates:

            # find a function that builds _source
            func = self.__sources_base_ep_func[key]
            if not func:
                continue

            src = func()

            mqt_ep = self.__get_modify_query_times(key, start, end, sources_start, sources_end)
            if mqt_ep:
                self.__sources_modify_query_times[key] = True
                src.sink(mqt_ep)

            if need_to_bind_symbol:
                bound = None
                if key in self.__sources_symbols:  # TODO: this is wrong, we need to know about symbols
                    # it happens when we do not copy symbols when apply
                    # merge with bound symbols.
                    # Wrong, in that case merge with bound symbol is
                    # non distinguishable from the manually passed None
                    # for external queries
                    bound = self.__sources_symbols[key]
                    if isinstance(bound, _ManuallyBoundValue):
                        bound = bound.value

                if bound and bound is not adaptive and bound is not adaptive_to_default:
                    src.__node.symbol(bound)
                else:
                    # if key is not in __sources_symbols, then
                    # it means that symbol was not specified, and
                    # therefor use unbound symbol
                    if common_symbol is None:
                        if bound is adaptive_to_default:
                            src.__node.symbol(configuration.config.default_symbol)
                        # TODO: write test validated this
                        # else:
                        #     raise Exception("One of the branch does not have symbol specified")

            # --------------------------
            # glue _source with the main graph
            self.node().add_rules(src.node().copy_rules())
            self.source_by_key(src.node().copy_graph(), key)
            self._merge_tmp_otq(src)

        if start is None:
            start = sources_start

        if end is None:
            end = sources_end

        return start, end, common_symbol

    def _to_graph(self, add_passthrough=True):
        """
        Construct the graph. Only for internal usage.

        It is private, because it constructs the raw graph assuming that a graph
        is already defined, and might confuse an end user, because by default Source
        is not fully defined; it becomes fully defined only when symbols, start and
        end datetime are specified.
        """
        constructed_obj = self.copy()

        # we add it for case when the last EP has a pin output
        if add_passthrough:
            constructed_obj.sink(otq.Passthrough())

        return otq.GraphQuery(constructed_obj.node().get())

    def to_graph(self, raw=None, symbols=None, start=None, end=None, *, add_passthrough=True):
        """
        Construct an :py:class:`onetick.query.GraphQuery` object.

        Parameters
        ----------
        raw:
            .. deprecated:: 1.4.17 has no effect

        symbols:
            symbols query to add to otq.GraphQuery
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            start time of a query
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            end time of a query
        add_passthrough: bool
            add additional :py:class:`onetick.query.Passthrough` event processor to the end of a resulted graph

        Returns
        -------
        otq.GraphQuery

        See Also
        --------
        :meth:`render`
        """

        if raw is not None:
            warnings.warn('The "raw" flag is deprecated and makes not effect', FutureWarning)

        _obj, _start, _end, _symbols = self.__prepare_graph(symbols, start, end)

        if _obj._tmp_otq.queries:
            warnings.warn('Using .to_graph() for a Source object that uses sub-queries! '
                          'This operation is deprecated and is not guaranteed to work as expected. '
                          'Such a Source should be executed using otp.run() or saved to disk using to_otq()',
                          FutureWarning)
            _obj.sink(otq.Passthrough().output_pin_name('OUT_FOR_TO_GRAPH'))
            _graph = _obj._to_graph(add_passthrough=False)
            _graph.set_start_time(_start)
            _graph.set_end_time(_end)
            _graph.set_symbols(_symbols)

            query = _obj._tmp_otq.save_to_file(query=_graph, file_suffix='_to_graph.otq')
            query_path, query_name = query.split('::')
            query_params = get_query_parameter_list(query_path, query_name)

            source_with_nested_query = otp.Query(otp.query(query,
                                                           **{param: f'${param}' for param in query_params}),
                                                 out_pin='OUT_FOR_TO_GRAPH')
            return source_with_nested_query.to_graph(
                symbols=_symbols, start=_start, end=_end,
                add_passthrough=add_passthrough)
        else:
            return _obj._to_graph(add_passthrough=add_passthrough)

    def render(self, **kwargs):
        """
        Renders a calculation graph using the ``graphviz`` library.
        Every node is the onetick query language event processor.
        Nodes in nested queries, first stage queries and eval queries are not shown.
        Could be useful for debugging and in jupyter to learn the underlying graph.

        Note that it's required to have :graphviz:`graphviz <>` package installed.

        Examples
        --------
        >>> data = otp.Tick(X=3)
        >>> data1, data2 = data[(data['X'] > 2)]
        >>> data = otp.merge([data1, data2])
        >>> data.render()  # doctest: +SKIP

        .. graphviz:: ../../static/render_example.dot
        """
        kwargs.setdefault('verbose', True)
        self.to_graph().render(**kwargs)

    def render_otq(
        self,
        image_path: Optional[str] = None,
        output_format: Optional[str] = None,
        load_external_otqs: bool = True,
        view: bool = False,
        line_limit: Optional[Tuple[int, int]] = (10, 30),
        parse_eval_from_params: bool = False,
        render_debug_info: bool = False,
        debug: bool = False,
        graphviz_compat_mode: bool = False,
        font_family: Optional[str] = None,
        font_size: Optional[Union[int, float]] = None,
        **kwargs,
    ):
        """
        Render current :py:class:`~onetick.py.Source` graph.

        Parameters
        ----------
        image_path: str, None
            Path for generated image. If omitted, image will be saved in a temp dir
        output_format: str, None
            `Graphviz` rendering format. Default: `png`.
            If `image_path` contains one of next extensions, `output_format` will be set automatically:
            `png`, `svg`, `dot`.
        load_external_otqs: bool
            If set to `True` (default) dependencies from external .otq files (not listed in ``path`` param)
            will be loaded automatically.
        view: bool
            Defines should generated image be shown after render.
        line_limit: Tuple[int, int], None
            Limit for maximum number of lines and length of some EP parameters strings.
            First param is limit of lines, second - limit of characters in each line.
            If set to None limit disabled.
            If one of tuple values set to zero the corresponding limit disabled.
        parse_eval_from_params: bool
            Enable parsing and printing `eval` sub-queries from EP parameters.
        render_debug_info: bool
            Render additional debug information.
        debug: bool
            Allow to print stdout or stderr from `Graphviz` render.
        graphviz_compat_mode: bool
            Change internal parameters of result graph for better compatibility with old `Graphviz` versions.
            Could produce larger and less readable graphs.
        font_family: str, optional
            Font name

            Default: **Monospace**
        font_size: int, float, str, optional
            Font size
        kwargs:
            Additional arguments to be passed to :py:meth:`onetick.py.Source.to_otq` method (except
            ``file_name``, ``file_suffix`` and ``query_name`` parameters)

        Returns
        -------
        Path to rendered image

        See also
        --------
        :py:func:`render_otq <onetick.py.utils.render_otq>`

        Examples
        --------

        >>> data = otp.DataSource(db='US_COMP', tick_type='TRD', symbols='AAA')  # doctest: +SKIP
        >>> data1, data2 = data[(data['PRICE'] > 50)]  # doctest: +SKIP
        >>> data = otp.merge([data1, data2])  # doctest: +SKIP
        >>> data.render_otq('./path/to/image.png')  # doctest: +SKIP

        .. image:: ../../static/testing/images/render_otq_3.png
        """

        if {'file_name', 'file_suffix', 'query_name'} & kwargs.keys():
            raise ValueError(
                'It\'s not allowed to pass parameters `file_name`, `file_suffix` and `query_name` as `kwargs` '
                'in `render_otq` method.'
            )

        otq_path = self.to_otq(**kwargs)
        return render_otq(
            otq_path, image_path, output_format, load_external_otqs, view, line_limit, parse_eval_from_params,
            render_debug_info, debug, graphviz_compat_mode, font_family, font_size,
        )

    def copy(self, ep=None, columns=None, deep=False) -> 'Source':
        """
        Build an object with copied calculation graph.

        Every node of the resulting graph has the same id as in the original. It means that
        if the original and copied graphs are merged or joined together further then all common
        nodes (all that created before the .copy() method) will be glued.

        For example, let's imagine that you have the following calculation graph ``G``

        .. graphviz::

           digraph {
             rankdir="LR";
             A -> B;
           }

        where ``A`` is a source and ``B`` is some operation on it.

        Then we copy it to the ``G'`` and assign a new operation there

        .. graphviz::

           digraph {
             rankdir="LR";
             A -> B -> C;
           }

        After that we decided to merge ``G`` and ``G'``. The resulting calculation graph will be:

        .. graphviz::

           digraph {
             rankdir="LR";
             A -> B -> C -> MERGE;
             B -> MERGE;
           }

        Please use the :meth:`Source.deepcopy` if you want to get the following calculation graph after merges and joins

        .. graphviz::

           digraph {
             rankdir="LR";
             A -> B -> C -> MERGE;
             "A'" -> "B'" -> "C'" -> MERGE;
           }

        Returns
        -------
        Source

        See Also
        --------
        Source.deepcopy
        """
        if columns is None:
            columns = self.columns(skip_meta_fields=True)

        if ep:
            result = self.__class__(node=ep, schema=columns)
            result.source(self.node().copy_graph())
            # we need to clean it, because ep is not a _source
            result._clean_sources_dates()
        else:
            result = self.__class__(node=self.node(), schema=columns)

        result.node().add_rules(self.node().copy_rules(deep=deep))
        result._set_sources_dates(self)
        if deep:
            # generating all new uuids for node history and for sources
            # after they were initialized
            keys = defaultdict(uuid.uuid4)  # type: ignore
            result.node().rebuild_graph(keys)
            result._change_sources_keys(keys)

        # add state
        result._copy_state_vars_from(self)

        result._tmp_otq = self._tmp_otq.copy()
        result.__name = self.__name     #noqa

        result._copy_properties_from(self)

        return result

    def deepcopy(self, ep=None, columns=None) -> 'onetick.py.Source':
        """
        Copy all graph and change ids for every node.
        More details could be found in :meth:`Source.copy`

        See Also
        --------
        Source.copy
        """
        return self.copy(ep, columns, deep=True)

    def _copy_properties_from(self, obj):
        # needed if we are doing copy of a child with custom properties
        for attr in set(self.__class__._PROPERTIES) - set(Source._PROPERTIES):
            setattr(self, attr, getattr(obj, attr))

    def _copy_state_vars_from(self, objs):
        self.__dict__["_state_vars"] = StateVars(self, objs)

    def columns(self, skip_meta_fields=False):
        """
        Return columns in data source

        Parameters
        ----------
        skip_meta_fields: bool, default=False
            do not add meta fields
        Returns
        -------
        dict
        """
        result = {}

        for key, value in self.__dict__.items():
            if skip_meta_fields and self._check_key_is_meta(key):
                continue

            if self._check_key_in_properties(key):
                continue

            if isinstance(value, _Column):
                result[value.name] = value.dtype

        return result

    def drop_columns(self):
        """
        Method removes all columns in the python representation, but don't
        drop columns on the data.

        It is used when external query is applied, because we don't know how
        data schema has changed.
        """

        items = []

        for key, value in self.__dict__.items():
            if self._check_key_is_reserved(key):
                continue

            if isinstance(value, _Column):
                items.append(key)

        for item in items:
            del self.__dict__[item]

    def node(self):
        return self.__node

    def tick_type(self, tt):
        self.__node.tick_type(tt)
        return self

    def symbol(self, symbol):  # NOSONAR
        """
        Apply symbol to graph

        .. deprecated:: 1.3.31

        """
        warnings.warn("symbol method is deprecated, please specify symbol during creation", FutureWarning)
        self.__node.symbol(symbol)
        return self

    def node_name(self, name=None, key=None):
        return self.__node.node_name(name, key)

    def _fix_varstrings(self):
        """
        PY-556: converting to varstring results in string with null-characters
        """
        varstring_columns = {
            name: self[name]
            for name, dtype in self.schema.items()
            if dtype is ott.varstring
        }
        # just updating the column removes null-characters
        if varstring_columns:
            self.update(varstring_columns, inplace=True)

    def __from_ep_to_proxy(self, ep):
        in_pin, out_pin = None, None
        if isinstance(ep, otq.graph_components.EpBase.PinnedEp):
            if hasattr(ep, "_output_name"):
                out_pin = getattr(ep, "_output_name")
            else:
                in_pin = getattr(ep, "_input_name")

            ep = getattr(ep, "_ep")

        return ep, uuid.uuid4(), in_pin, out_pin

    def sink(self, ep, out_pin=None, inplace: bool = True):
        """
        Appends ``ep`` node to this source (inplace by default).
        Connects ``out_pin`` of this source to ``ep``.

        Can be used to connect onetick.query objects to :class:`onetick.py.Source`.

        Data schema changes (added or deleted columns) will not be detected automatically
        after applying this function, so the user must change the schema himself
        by updating :meth:`onetick.py.Source.schema` property.

        Parameters
        ----------
        ep: otq.graph_components.EpBase,\
            otq.graph_components.EpBase.PinnedEp,\
            Tuple[otq.graph_components.EpBase, uuid.uuid4, Optional[str], Optional[str]]
            onetick.query EP object to append to source.
        out_pin: Optional[str], default=None
            name of the out pin to connect to ``ep``
        inplace: bool, default=False
            if `True` method will modify current object,
            otherwise it will return modified copy of the object.

        Returns
        ----------
        :class:`Source` or ``None``
            Returns ``None`` if ``inplace=True``.

        See Also
        --------
        onetick.py.Source.schema
        onetick.py.core._source.schema.Schema

        Examples
        --------
        Adding column 'B' directly with onetick.query EP.

        >>> data = otp.Tick(A=1)
        >>> data.sink(otq.AddField(field='B', value=2)) # OTdirective: skip-snippet:;
        >>> otp.run(data) # OTdirective: skip-snippet:;
                Time  A  B
        0 2003-12-01  1  2

        But we can't use this column with `onetick.py` methods yet:

        >>> data['C'] = data['B'] # OTdirective: skip-snippet:; # doctest: +ELLIPSIS
        Traceback (most recent call last):
         ...
        AttributeError: There is no 'B' column

        We should manually change source's schema:

        >>> data.schema.update(B=int) # OTdirective: skip-snippet:;
        >>> data['C'] = data['B']
        >>> otp.run(data)
                Time  A  B  C
        0 2003-12-01  1  2  2

        Use parameter ``inplace=False`` to return modified copy of the source:

        >>> data = otp.Tick(A=1)
        >>> new_data = data.sink(otq.AddField(field='B', value=2), inplace=False)
        >>> otp.run(data)
                Time  A
        0 2003-12-01  1
        >>> otp.run(new_data)
                Time  A  B
        0 2003-12-01  1  2
        """
        if not (
            issubclass(type(ep), otq.graph_components.EpBase)
            or issubclass(type(ep), otq.graph_components.EpBase.PinnedEp)
            or isinstance(ep, tuple)
        ):
            raise TypeError("sinking is allowed only for EpBase instances")

        if inplace:
            obj = self
        else:
            obj = self.copy()

        if isinstance(ep, tuple):
            # for already existed EP fetched from _ProxyNode
            obj.__node.sink(out_pin, *ep)
        else:
            obj.__node.sink(out_pin, *obj.__from_ep_to_proxy(ep))

        if inplace:
            return None
        return obj

    def __rshift__(self, ep):
        """ duplicates sink() method, but returns new object """
        new_source = self.copy()
        new_source.sink(ep)
        return new_source

    def __irshift__(self, ep):
        """ duplicates sink() method, but assigns source new object """
        new_source = self.copy()
        new_source.sink(ep)
        return new_source

    def source(self, ep, in_pin=None):
        """ Add node as source to root node """
        if not (
            issubclass(type(ep), otq.graph_components.EpBase)
            or issubclass(type(ep), otq.graph_components.EpBase.PinnedEp)
            or isinstance(ep, tuple)
        ):
            raise TypeError("sourcing is allowed only for EpBase instances")

        if isinstance(ep, tuple):
            # for already existed EP fetched from _ProxyNode
            return self.__node.source(in_pin, *ep)
        else:
            return self.__node.source(in_pin, *self.__from_ep_to_proxy(ep))

    def source_by_key(self, ep, to_key):
        """ Add node as source to graph node by key"""
        if not (
            issubclass(type(ep), otq.graph_components.EpBase)
            or issubclass(type(ep), otq.graph_components.EpBase.PinnedEp)
            or isinstance(ep, tuple)
        ):
            raise TypeError("sourcing is allowed only for EpBase instances")

        if isinstance(ep, tuple):
            # for already existed EP fetched from _ProxyNode
            return self.__node.source_by_key(to_key, *ep)
        else:
            return self.__node.source_by_key(to_key, *self.__from_ep_to_proxy(ep))

    def to_symbol_param(self):
        """
        Creates a read-only instance with the same columns except Time.
        It is used as a result of a first stage query with symbol params.

        See also
        --------
        :ref:`static/concepts/symbols:Symbol parameters`

        Examples
        --------
        >>> symbols = otp.Ticks({'SYMBOL_NAME': ['S1', 'S2'], 'PARAM': ['A', 'B']})
        >>> symbol_params = symbols.to_symbol_param()
        >>> t = otp.DataSource('SOME_DB', tick_type='TT')
        >>> t['S_PARAM'] = symbol_params['PARAM']
        >>> result = otp.run(t, symbols=symbols)
        >>> result['S1']
                             Time  X S_PARAM
        0 2003-12-01 00:00:00.000  1       A
        1 2003-12-01 00:00:00.001  2       A
        2 2003-12-01 00:00:00.002  3       A
        """
        return _SymbolParamSource(**self.columns())

    @staticmethod
    def _convert_symbol_to_string(symbol, tmp_otq=None, start=None, end=None, timezone=None, symbol_date=None):
        if start is adaptive:
            start = None
        if end is adaptive:
            end = None

        if isinstance(symbol, Source):
            symbol = otp.eval(symbol).to_eval_string(tmp_otq=tmp_otq,
                                                     start=start, end=end, timezone=timezone,
                                                     operation_suffix='symbol',
                                                     query_name=None,
                                                     file_suffix=symbol._name_suffix('symbol.otq'),
                                                     symbol_date=symbol_date)

        if isinstance(symbol, otp.query):
            return symbol.to_eval_string()

        if isinstance(symbol, otq.GraphQuery):
            params = {'symbol_date': symbol_date} if symbol_date is not None else {}
            query_name = tmp_otq.add_query(symbol, suffix='__symbol', params=params)
            return f'eval(THIS::{query_name})'

        return symbol

    @staticmethod
    def _construct_multi_branch_graph(branches):
        # TODO: add various checks, e.g. that branches have common parts
        main = branches[0].copy()
        for branch in branches[1:]:
            main.node().add_rules(branch.node().copy_rules())
            main._merge_tmp_otq(branch)
        return main

    def _apply_side_branches(self, side_branches):
        for side_branch in side_branches:
            self.node().add_rules(side_branch.node().copy_rules())
            self._merge_tmp_otq(side_branch)
            self.__sources_keys_dates.update(side_branch.__sources_keys_dates)
            self.__sources_modify_query_times.update(side_branch.__sources_modify_query_times)
            self.__sources_base_ep_func.update(side_branch.__sources_base_ep_func)
            self.__sources_symbols.update(side_branch.__sources_symbols)

    @property
    def state_vars(self) -> StateVars:
        """
        Provides access to state variables

        Returns
        -------
        State Variables: Dict[str, state variable]
            State variables, you can access one with its name.

        See Also
        --------
        | `State Variables \
         <../../static/getting_started/variables_and_data_structures.html#variables-and-data-structures>`_
        | **DECLARE_STATE_VARIABLES** OneTick event processor

        """
        return self.__dict__['_state_vars']

    # non word characters are not supported
    __invalid_query_name_symbols_regex = re.compile(r'\W')

    def __remove_invalid_symbols(self, s):
        """
        Replaces symbols that cannot be put in query names with '_'
        """
        return self.__invalid_query_name_symbols_regex.sub('_', s)

    def get_name(self, remove_invalid_symbols=False) -> Optional[str]:
        """
        Returns source name.

        Parameters
        ----------
        remove_invalid_symbols: bool
            If True, all characters not supported in query names in `.otq` file will be replaced,
            because only alphanumeric, minus and underscore characters are supported in query names.

        See also
        --------
        :meth:`set_name`
        """
        if remove_invalid_symbols and self.__name:
            return self.__remove_invalid_symbols(self.__name)
        else:
            return self.__name

    def set_name(self, new_name):
        """
        Sets source name.
        It's an internal onetick-py name of the source that is only used
        as a part of the resulting .otq file name and as the name of the query inside this file.

        This method doesn't set the name of the OneTick graph node.

        Parameters
        ----------
        new_name: str
            New name of the source.

            Only alphanumeric, minus and underscore characters are supported.
            All other characters will be replaced in the resulting query name.

        See also
        --------
        :meth:`get_name`

        Examples
        --------
        >>> t = otp.Tick(A=1)

        By default source has no name and some predefined values are used when generating .otq file:

        >>> t.to_otq()  # doctest: +SKIP
        '/tmp/test_user/run_20240126_152546_1391/magnificent-wolverine.to_otq.otq::query'

        Changed name will be used as a part of the resulting .otq file name
        and as the name of the query inside this file:

        >>> t.set_name('main')
        >>> t.to_otq()  # doctest: +SKIP
        '/tmp/test_user/run_20240126_152546_1391/dandelion-angelfish.main.to_otq.otq::main'
        """
        assert isinstance(new_name, str) or new_name is None, "Source name must be a string or None."
        if new_name is not None:
            assert new_name != '', "Source name must be a non-empty string."
        self.__name = new_name

    def _name_suffix(self, suffix, separator='.', remove_invalid_symbols=False):
        if remove_invalid_symbols:
            suffix = self.__remove_invalid_symbols(suffix)
            separator = self.__remove_invalid_symbols(separator)
            name = self.get_name(remove_invalid_symbols=True)
        else:
            name = self.__name
        return f'{separator}{name}{separator}{suffix}' if name else f'{separator}{suffix}'

    @property
    def schema(self) -> Schema:
        """
        Represents actual python data schema in the column-name -> type format.
        For example, could be used after the :meth:`Source.sink` to adjust
        the schema.

        Returns
        -------
        Schema

        See Also
        --------
        Source.sink

        Examples
        --------

        >>> data = otp.Ticks([['X', 'Y',   'Z'],
        ...                   [  1, 0.5, 'abc']])
        >>> data['T'] = data['Time']
        >>> data.schema
        {'X': <class 'int'>, 'Y': <class 'float'>, 'Z': <class 'str'>, 'T': <class 'onetick.py.types.nsectime'>}

        >>> data.schema['X']
        <class 'int'>

        >>> data.schema['X'] = float
        >>> data.schema['X']
        <class 'float'>

        >>> 'W' in data.schema
        False
        >>> data.schema['W'] = otp.nsectime
        >>> 'W' in data.schema
        True
        >>> data.schema['W']
        <class 'onetick.py.types.nsectime'>
        """
        schema = self.columns(skip_meta_fields=True)
        # meta fields will be in schema, but hidden
        hidden_columns = {
            k: v
            for k, v in self.columns(skip_meta_fields=False).items()
            if self._check_key_is_meta(k)
        }
        if 'TIMESTAMP' in hidden_columns:
            hidden_columns['Time'] = hidden_columns['TIMESTAMP']
        return Schema(_base_source=self, _hidden_columns=hidden_columns, **schema)

    def set_schema(self, **kwargs):
        """
        Set schema of the source.
        Note: this method affect python part only and won't make any db queries. It used to set schema after db reading/
        complex query.

        .. deprecated:: 1.14.9

        Please use the :property:`Source.schema` to access and adjust the schema.

        Parameters
        ----------
        kwargs
            schema in the column_name=type format

        Examples
        --------
        Python can't follow low level change of column, e.g. complex query or pertick script can be sink.

        >>> data = otp.Ticks(dict(A=[1, 2], B=["a", "b"]))
        >>> data.sink(otq.AddField(field='Z', value='5'))
        >>> data.columns(skip_meta_fields=True)
        {'A': <class 'int'>, 'B': <class 'str'>}
        >>> # OTdirective: snippet-name: Arrange.schema.set;
        >>> data.set_schema(A=int, B=str, Z=int)
        >>> data.columns(skip_meta_fields=True)
        {'A': <class 'int'>, 'B': <class 'str'>, 'Z': <class 'int'>}
        """
        self.drop_columns()
        for name, dtype in kwargs.items():
            dtype = ott.get_source_base_type(dtype)
            if self._check_key_is_meta(name):
                warnings.warn(f"Setting type in schema for meta field {name}", stacklevel=2)
            if self._check_key_in_properties(name):
                raise ValueError(f"Can't set type in schema for class property {name}")
            self.__dict__[name] = _Column(name, dtype, self)

    def has_start_end_time(self) -> Tuple[bool, bool]:
        """
        Check if at least one of query sources has start and end time
        """
        has_start_time = False
        has_end_time = False

        for start, end in self._get_sources_dates().values():
            if not has_start_time and start is not adaptive and start is not None:
                has_start_time = True

            if not has_end_time and end is not adaptive and end is not None:
                has_end_time = True

        return has_start_time, has_end_time

    from ._source.source_methods.aggregations import (  # type: ignore[misc]
        agg,
        high, low, first, last, distinct, high_time, low_time,
        ob_snapshot, ob_snapshot_wide, ob_snapshot_flat, ob_summary,
        ob_size, ob_vwap, ob_num_levels,
        ranking, percentile, find_value_for_percentile,
        exp_w_average, exp_tw_average, standardized_moment,
        portfolio_price, multi_portfolio_price, return_ep, implied_vol,
        linear_regression, partition_evenly_into_groups,
        process_by_group,
    )
    from ._source.source_methods.joins import (  # type: ignore[misc]
        _process_keep_time_param,
        _get_columns_with_prefix,
        join_with_collection,
        join_with_query,
        point_in_time,
        join_with_snapshot,
    )
    from ._source.source_methods.times import (  # type: ignore[misc]
        update_timestamp,
        modify_query_times,
        time_interval_shift,
        time_interval_change,
    )
    from ._source.source_methods.fields import (  # type: ignore[misc]
        _add_field,
        _update_timestamp,
        _update_field,
        __setattr__,
        __setitem__,
        add_fields,
        table,
        update,
    )
    from ._source.source_methods.filters import (  # type: ignore[misc]
        if_else,
        where_clause,
        where,
        _get_integer_slice,
        __getitem__,
        dropna,
        time_filter,
        skip_bad_tick,
        character_present,
        primary_exch,
    )
    from ._source.source_methods.drops import (  # type: ignore[misc]
        drop,
        __delitem__,
    )
    from ._source.source_methods.writes import (  # type: ignore[misc]
        write,
        write_parquet,
        save_snapshot,
        write_text,
    )
    from ._source.source_methods.renames import (  # type: ignore[misc]
        _add_prefix_and_suffix,
        add_prefix,
        add_suffix,
        rename,
    )
    from ._source.source_methods.pandases import (  # type: ignore[misc]
        plot,
        count,
        head,
        tail,
    )
    from ._source.source_methods.sorts import (  # type: ignore[misc]
        sort_values,
        sort,
    )
    from ._source.source_methods.debugs import (  # type: ignore[misc]
        dump,
        throw,
        logf,
    )
    from ._source.source_methods.applyers import (  # type: ignore[misc]
        apply_query,
        apply,
        script,
    )
    from ._source.source_methods.symbols import (  # type: ignore[misc]
        show_symbol_name_in_db,
        modify_symbol_name,
    )
    from ._source.source_methods.columns import (  # type: ignore[misc]
        mean,
        unite_columns,
    )
    from ._source.source_methods.switches import (  # type: ignore[misc]
        switch,
        split,
    )
    from ._source.source_methods.misc import (  # type: ignore[misc]
        pause,
        insert_tick,
        insert_at_end,
        transpose,
        cache,
        pnl_realized,
        execute,
        _columns_names_regex,
        fillna,
        mkt_activity,
        book_diff,
        limit,
        virtual_ob,
        corp_actions,
    )
    from ._source.source_methods.merges import (  # type: ignore[misc]
        __add__,
        append,
        diff,
        lee_and_ready,
        estimate_ts_delay,
    )
    from ._source.source_methods.data_quality import (  # type: ignore[misc]
        show_data_quality,
        insert_data_quality_event,
        intercept_data_quality,
        show_symbol_errors,
        intercept_symbol_errors,
    )


_Source = Source    # Backward compatibility



import datetime
import inspect

import pandas as pd
import numpy as np

from onetick.py import types as ott


def are_numerics(*dtypes):
    return all(
        inspect.isclass(dtype)
        and (issubclass(dtype, (float, int)) or np.issubdtype(dtype, np.integer) or issubclass(dtype, ott.decimal))
        and not issubclass(dtype, (ott.nsectime, ott.msectime))
        for dtype in dtypes
    )


def are_ints_not_time(*dtypes):
    return all(inspect.isclass(dtype)
               and (issubclass(dtype, int) or np.issubdtype(dtype, np.integer))
               and not issubclass(dtype, (ott.nsectime, ott.msectime)) for dtype in dtypes)


def are_time(*dtypes):
    return all(inspect.isclass(dtype) and issubclass(dtype, (ott.nsectime, ott.msectime)) for dtype in dtypes)


def are_ints_or_time(*dtypes):
    return all(inspect.isclass(dtype)
               and (issubclass(dtype, (int, datetime.datetime, ott.datetime, pd.Timestamp))
                    or np.issubdtype(dtype, np.integer))
               for dtype in dtypes)


def are_floats(*dtypes):
    return all(inspect.isclass(dtype) and issubclass(dtype, (float, np.float64, np.double)) for dtype in dtypes)


def are_strings(*dtypes):
    return all(inspect.isclass(dtype) and issubclass(dtype, str) for dtype in dtypes)


def are_bools(*dtypes):
    return all(inspect.isclass(dtype) and issubclass(dtype, bool) for dtype in dtypes)


def _get_widest_type(left, right):
    """ return the widest type

    If the left is subclass of right return left, if right is subclass of left return right,
        returns None in all other cases. ott.nsectime, ott.msectime and bool won't be considered as int, only
        custom classes.

    Parameters
    ----------
    left: type
        type of the first argument
    right: type
        type of the second argument

    Returns
    -------
    result: type or None
        widest type if it exist, or None in another case

    Examples
    --------

    >>> class MyInt(int):   # OTdirective: skip-snippet:;
    ...     pass
    >>> class MyFloat(float):
    ...     pass
    >>> def test():
    ...     print(_get_widest_type(int, MyInt))
    ...     print(_get_widest_type(float, MyFloat))
    ...     print(_get_widest_type(int, float))
    ...     print(_get_widest_type(MyInt, MyFloat))
    ...     print(_get_widest_type(bool, float))
    ...     print(_get_widest_type(int, bool))
    ...     print(_get_widest_type(MyInt, bool))
    ...     print(_get_widest_type(bool, MyFloat))
    >>> test()  # doctest: +ELLIPSIS
    <class '...MyInt'>
    <class '...MyFloat'>
    <class 'float'>
    <class '...MyFloat'>
    <class 'float'>
    <class 'int'>
    <class '...MyInt'>
    <class '...MyFloat'>

    >>> from onetick.py import types as ott     # OTdirective: skip-snippet:;
    >>> class MyString(str):
    ...     pass
    >>> (_get_widest_type(str, MyString),
    ...  _get_widest_type(str, ott.string[15]),
    ...  _get_widest_type(MyString, ott.string[15]))  # doctest: +ELLIPSIS
    (<class 'onetick.py.core.column_operations._methods.op_types.MyString'>, string[15], None)

    >>> from onetick.py import types as ott     # OTdirective: skip-snippet:;
    >>> _get_widest_type(int, ott.msectime), _get_widest_type(int, ott.nsectime)
    (None, None)

    >>> from onetick.py import types as ott     # OTdirective: skip-snippet:;
    >>> class MyTime(ott.OTPBaseTimeStamp):
    ...     pass
    >>> class MyNSec(ott.nsectime):
    ...     pass
    >>> (_get_widest_type(MyTime, ott.OTPBaseTimeStamp), _get_widest_type(MyTime, ott.nsectime),
    ... _get_widest_type(MyNSec, ott.nsectime))  # doctest: +ELLIPSIS
    (<class '...MyTime'>, None, <class '...MyNSec'>)
    """

    # decimal takes precedence before integer and floating point types
    if issubclass(left, ott.decimal) and are_numerics(right):
        return left
    if are_numerics(left) and issubclass(right, ott.decimal):
        return right

    if issubclass(left, float) and issubclass(right, float):
        # between np.float and float we choose base float
        if left is not float and np.issubdtype(left, np.floating):
            left = float
        if right is not float and np.issubdtype(right, np.floating):
            right = float

    if issubclass(left, float):
        if are_ints_not_time(right):
            return left
        else:
            return _get_wideclass(left, right) if issubclass(right, float) else None
    elif issubclass(right, float):
        if are_ints_not_time(left):
            return right
        else:
            return _get_wideclass(left, right) if issubclass(left, float) else None

    if are_time(left) and are_ints_not_time(right) or are_time(right) and are_ints_not_time(left):
        return None

    if issubclass(right, int) and issubclass(left, bool):
        return right
    if issubclass(left, int) and issubclass(right, bool):
        return left

    if issubclass(right, int) and np.issubdtype(left, np.integer):
        return right
    if issubclass(left, int) and np.issubdtype(right, np.integer):
        return left

    return _get_wideclass(left, right)


def _get_wideclass(left, right):
    if issubclass(left, right):
        return left
    if issubclass(right, left):
        return right
    return None



from typing import Union

import onetick.py as otp
from onetick.py.otq import otq

from onetick.py.core.source import Source

from .. import types as ott
from .. import utils

from .common import update_node_tick_type


def process_otq_params(params: Union[dict, None]) -> str:
    if not params:
        return ""

    return ",".join([f"{k}={ott.value2str(v)}" for k, v in params.items()])


class ReadCache(Source):
    def __init__(
        self,
        cache_name=None,
        db=utils.adaptive_to_default,
        symbol=utils.adaptive,
        tick_type=utils.adaptive,
        start=utils.adaptive,
        end=utils.adaptive,
        read_mode="automatic",
        update_cache=True,
        otq_params=None,
        create_cache_query="",
        schema=None,
        **kwargs,
    ):
        """
        Make cached query

        Cache is initialized on the first read attempt.

        Parameters
        ----------
        cache_name: str
            Name of cache for a query.
        symbol: str, list of str, :class:`Source`, :class:`query`, :py:func:`eval query <onetick.py.eval>`
            Symbol(s) from which data should be taken.
        db: str
            Database to use for tick generation.
        tick_type: str
            Tick type.
            Default: ANY.
        start: :py:class:`otp.datetime <onetick.py.datetime>`
            Start time for tick generation. By default the start time of the query will be used.
        end: :py:class:`otp.datetime <onetick.py.datetime>`
            End time for tick generation. By default the end time of the query will be used.
        read_mode: str
            Mode of querying cache. One of these:

            * ``cache_only`` - only cached results are returned and queries are not performed.
            * ``query_only`` - the query is run irrespective of whether the result is already available in the cache.
            * ``automatic`` (default) - perform the query if the data is not found in the cache.
        update_cache: bool
            If set to ``True``, updates the cached data if ``read_mode=query_only`` or if ``read_mode=automatic`` and
            the result data not found in the cache. Otherwise, the cache remains unchanged.
        otq_params: dict
            OTQ parameters to override ``otq_params`` that are passed during creation of cache.
            Query result will be cached separately for each unique pair of OTQ parameters.
        create_cache_query: str
            If a cache with the given name is not present, the query provided in this param will be invoked,
            which should contain CREATE_CACHE EP to create the corresponding cache.
        schema: Optional, dict
            Dictionary of columns names with their types.
        kwargs:
            Deprecated. Use ``schema`` instead.
            Dictionary of columns names with their types.

        See also
        --------
        | **READ_CACHE** OneTick event processor
        | :py:func:`onetick.py.create_cache`

        Examples
        --------
        Simple cache read:

        >>> def query_func():
        ...    return otp.DataSource("DEMO_L1", tick_type="TRD", symbols="AAPL")
        >>> otp.create_cache(
        ...    cache_name="some_cache", query=query_func, tick_type="TRD", db="DEMO_L1",
        ... )
        >>> src = otp.ReadCache("some_cache")
        >>> otp.run(src)  # doctest: +SKIP

        Make cache query for specific time interval:

        >>> src = otp.ReadCache(
        ...    "some_cache",
        ...    start=otp.datetime(2024, 1, 1, 12, 30),
        ...    end=otp.datetime(2024, 1, 2, 18, 0),
        ... )
        >>> otp.run(src)  # doctest: +SKIP

        Override or set `otq_params` for one query:

        >>> src = otp.ReadCache(
        ...    "some_cache",
        ...    otq_params={"some_param": "test_value"},
        ... )
        >>> otp.run(src)  # doctest: +SKIP
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if cache_name is None:
            raise ValueError("Missing required parameter `cache_name`")

        if read_mode not in {"automatic", "cache_only", "query_only"}:
            raise ValueError(f"Incorrect value of `read_mode` parameter passed: {read_mode}")

        if update_cache and read_mode == 'cache_only':
            raise ValueError("`update_cache` parameter couldn't be set to True with `read_mode` set to cache_only")

        read_mode = read_mode.upper()

        otq_params = process_otq_params(otq_params)

        super().__init__(
            _symbols=symbol,
            _start=start,
            _end=end,
            _base_ep_func=lambda: self.base_ep(
                db=db,
                tick_type=tick_type,
                cache_name=cache_name,
                read_mode=read_mode,
                update_cache=update_cache,
                otq_params=otq_params,
                create_cache_query=create_cache_query,
            ),
            schema=schema,
            **kwargs,
        )

    def base_ep(
        self,
        db=utils.adaptive_to_default,
        tick_type=utils.adaptive,
        cache_name=None,
        read_mode="automatic",
        update_cache=True,
        otq_params="",
        create_cache_query="",
    ):
        src = Source(
            otq.ReadCache(
                cache_name=cache_name,
                per_cache_otq_params=otq_params,
                read_mode=read_mode,
                update_cache=update_cache,
                create_cache_query=create_cache_query,
            )
        )

        update_node_tick_type(src, tick_type, db)

        return src



from typing import TYPE_CHECKING
from copy import deepcopy

if TYPE_CHECKING:
    from onetick.py.core.source import Source   # hack for annotations

from onetick.py.otq import otq

from onetick.py.core.column import _Column
from onetick.py import types as ott

from ._base import _Aggregation
from ._docs import (_running_doc,
                    _all_fields_doc,
                    _bucket_interval_doc,
                    _bucket_time_doc,
                    _bucket_units_doc,
                    _bucket_end_condition_doc,
                    _boundary_tick_bucket_doc,
                    _group_by_doc,
                    _groups_to_display_doc)
from onetick.py.docs.utils import docstring


# OneTick build >= 20220913120000
if hasattr(otq, 'NumDistinct'):

    class NumDistinct(_Aggregation):
        NAME = 'NUM_DISTINCT'
        EP = otq.NumDistinct

        FIELDS_MAPPING = deepcopy(_Aggregation.FIELDS_MAPPING)
        FIELDS_MAPPING['keys'] = 'KEYS'

        FIELDS_TO_SKIP = ['column_name', 'end_condition_per_group']

        output_field_type = int

        def __init__(self, keys, *args, **kwargs):
            super().__init__(column=_Column('TIMESTAMP'), *args, **kwargs)
            if isinstance(keys, str):
                keys = [keys]
            self._keys = keys

        @property
        def keys(self):
            return ott.value2str(','.join(self._keys))

        def apply(self, src, name='VALUE'):
            return super().apply(src, name)

        def validate_input_columns(self, src: 'Source'):
            for column in self._keys:
                if column not in src.schema:
                    raise TypeError(f"Aggregation {self.__class__.__name__} uses"
                                    f" column '{column}' as input, which doesn't exist")

    @docstring(parameters=[_running_doc, _all_fields_doc,
                           _bucket_interval_doc, _bucket_units_doc, _bucket_time_doc,
                           _bucket_end_condition_doc, _boundary_tick_bucket_doc, _group_by_doc, _groups_to_display_doc])
    def num_distinct(*args, **kwargs):
        """
        Outputs number of distinct values for a specified set of key fields.

        Parameters
        ----------
        keys: str or list of str or list of :py:class:`~onetick.py.Column`
            Specifies a list of tick attributes for which unique values are found.
            The ticks in the input time series must contain those attributes.

        Examples
        --------
        >>> data = otp.Ticks(dict(X=[1, 3, 2, 1, 3]))
        >>> data = data.agg({'X': otp.agg.num_distinct('X')})
        >>> otp.run(data)
                Time  X
        0 2003-12-04  3

        See also
        --------
        **NUM_DISTINCT** OneTick event processor
        """
        return NumDistinct(*args, **kwargs)



from typing import Optional, List

import onetick.py as otp
from onetick.py.otq import otq

import onetick.py.db._inspection
from onetick.py.core.source import Source

from .. import utils, configuration
from ..core.column_operations.base import _Operation
from ..compatibility import is_odbc_query_supported

from .common import update_node_tick_type


class ODBC(Source):

    _PROPERTIES = Source._PROPERTIES + [
        '_dsn',
        '_connection_string',
        '_authentication_type',
        '_user',
        '_password',
        '_sql',
        '_symbology',
        '_allow_unordered_ticks',
        '_tz',
        '_start_expr',
        '_end_expr',
        '_apply_symbol_name_history',
        '_numeric_scale',
        '_preserve_unicode_fields',
        '_db',
        '_tick_type',
        '_symbols',
        '_presort',
    ]

    def __init__(
        self,
        dsn: Optional[str] = None,
        connection_string: Optional[str] = None,
        authentication_type: str = 'credentials',
        user: Optional[str] = None,
        password: Optional[str] = None,
        sql: Optional[str] = None,
        start_expr: Optional[_Operation] = None,
        end_expr: Optional[_Operation] = None,
        tz=None,
        allow_unordered_ticks=False,
        numeric_scale=8,
        preserve_unicode_fields: Optional[List[str]] = None,
        symbology='',
        apply_symbol_name_history=False,
        start=utils.adaptive,
        end=utils.adaptive,
        db=utils.adaptive,
        tick_type=utils.adaptive,
        symbols=utils.adaptive_to_default,
        presort=False,
        schema=None,
        **kwargs,
    ):
        """
        Read from ODBC-compatible database and propagate the resulting data as a time series of ticks.

        A time series is generated for every symbol of the query and the symbol name can be passed down to SQL query.

        If an attribute with the name *TIMESTAMP* is present in the database schema,
        it is assumed to be of either SQL DATE or SQL TIMESTAMP type, in which case output tick timestamps
        will carry values of that field and the field itself will not be propagated.
        If such a field is absent, output tick timestamps will be equal to query end time.

        ODBC source supported only on OneTick versions starting from release 1.24
        (or from development build 20231108-0)

        Note
        ----

        To be able to use this class, you need to have a ODBC driver manager available on your machine
        (comes with Windows OS; **unixodbc** package may need to be installed on UNIX OS).

        Also, the following entry needs to be added to the main configuration file::

            LOAD_ODBC_UDF=true

        Parameters
        ----------
        dsn: str
            Target database's source name registered in ODBC configuration.
        connection_string: str
            ODBC connection string. The format depends on the database you are trying to connect.
            Example for SQLite database::

                DRIVER={SQLite3};Database=/path/to/the/database.db

            This parameter is mutually exclusive
            with parameters ``dsn``, ``user``, ``password`` and ``authentication_type``.
        authentication_type: str ('system' or 'credentials')

            - **system**: in this case a trusted connection to the database will be used.
              If authentication is enabled in the server configuration file
              (see the OneTick Installation and Administration Guide for details),
              the client will be authenticated and the client's credentials will be used for connection;
              otherwise, server credentials will be used.
            - **credentials**: username and password specified with parameter ``connection_string``
              or with parameters ``user`` and ``password`` will be used to connect to the database.

        user: str, optional
            The connection user name.

            Ignored if ``authentication_type`` is **system**.
        password: str, optional
            The connection password.

            Ignored if ``authentication_type`` is **system**.
        sql: str
            A query in SQL language, which may optionally contain parameter placeholders.

            There are 3 types of placeholders: **<_SYMBOL_NAME>**, **<_START_TIME>**, and **<_END_TIME>**.

            **<_SYMBOL_NAME>** will be replaced by the pure symbol name part of the query.

            **<_START_TIME>** and **<_END_TIME>** will be replaced by values of parameters
            ``start_expr`` and ``end_expr`` or, if they are not set, with values taken from query start and end times.

            These timestamps will be inserted in the format ``YYYY-MM-DD HH:MM:SS.sss TIMEZONE``.

        start_expr: str or :py:class:`~onetick.py.Operation`
            A constant string expression used to replace the **<_START_TIME>** placeholder in an ``sql`` query.
            Tick-dependent columns can't be used in this expression, only meta fields such as
            *TIMESTAMP*, *_START_TIME*, and *_END_TIME* are allowed.
        end_expr: str or :py:class:`~onetick.py.Operation`
            A constant string expression used to replace the **<_END_TIME>** placeholder in an ``sql`` query.
            Tick-dependent columns can't be used in this expression, only meta fields such as
            *TIMESTAMP*, *_START_TIME*, and *_END_TIME* are allowed.
        tz: str
            The timezone used to interpret *TIMESTAMP* and other datetime columns in the database.
            By default :py:attr:`tz<onetick.py.configuration.Config.tz>` is used.

        allow_unordered_ticks: bool
            If set to False, this class will raise an exception when unordered ticks will be encountered.

            If set to True, processing ticks unordered by timestamp will be allowed
            (exception may still be raised by other EPs of the graph that require ticks to be ordered).

        numeric_scale: int
            Number of digits after the decimal point for SQL_NUMERIC and SQL_DECIMAL data types.
            Precision (maximum number of digits) is always set to 34.
        preserve_unicode_fields: list of str
            By default ODBC queries all data as ANSI,
            ODBC Driver Manager then tries to convert Unicode characters to ANSI.

            This parameter is a list of fields, which have Unicode types in the data source
            and will be propagated without conversion.

            UNICODE_CHAR_TYPE field property will be set to UCS-2 for these fields.

        symbology: str
            If specified, **<_SYMBOL_NAME>** placeholder in ``sql`` query will be replaced with
            the mapping of the symbol name of the query for provided symbology.

        apply_symbol_name_history: bool
            If set to False, ODBC will not take into account the symbol name history
            when substituting the SQL query with actual symbol names.

            If set to True, ODBC will resolve symbol name changes according to the reference data
            by substituting the SQL query accordingly,
            provided the symbol date is specified in :py:func:`otp.run <onetick.py.run>` when running the query and
            the SQL query contains all three placeholders **<_SYMBOL_NAME>**, **<_START_TIME>**, **<_END_TIME>**.

            Note that ODBC data source needs to support "UNION ALL" syntax for this functionality to work.

        start:
            Custom start time of the query.
            By default the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        end:
            Custom end time of the query.
            By default the start time used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        db: str
            Custom database name for the node of the graph.
            By default the database used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        tick_type: str
            Custom tick type for the node of the graph.
            By default "ANY" tick type will be set.
        symbols: str or list of str
            Custom symbol name for the node of the graph or list of symbols.
            If list of symbols is specified, ticks from different symbols will be merged into one source.

            Separate ODBC connection will be created for each processed symbol.
            To avoid this and have a single connection per thread, ``presort`` parameter can be specified.

            By default the symbol name used by :py:func:`otp.run <onetick.py.run>` will be inherited.
        presort: bool
            Adds **PRESORT** EP before merging bound symbols specified in ``symbols``.
            That make ODBC use single connection for all symbols.
        schema: dict
            Set the schema of the python :py:class:`~onetick.py.Source` object of this class.

            Schema can't be taken automatically from the database, so it should be set manually
            for python-level type checking to work.
        kwargs:
            Deprecated. Use ``schema`` instead.
            Set the schema of the python :py:class:`~onetick.py.Source` object of this class.

        See also
        --------
        **ODBC_QUERY** OneTick event processor

        Examples
        --------

        Connect with database's ``dsn``, manually set schema, get all data from TEST_TABLE:

        >>> data = otp.ODBC(dsn='testdb_dsn', sql='select * from TEST_TABLE',
        ...                 schema={'A': str, 'B': int, 'C': float, 'D': otp.nsectime})  # doctest: +SKIP
        >>> otp.run(data)  # doctest: +SKIP
                Time   A     B        C                       D
        0 2003-12-04  A1  1975  8.12345 2022-01-01 12:13:14.111
        1 2003-12-04  A2  1971  7.98765 2022-01-02 22:23:24.222

        Connect using ``connection_string`` parameter:

        >>> data = otp.ODBC(connection_string='DRIVER={SQLite3};Database=/path/to/the/database',
        ...                 sql='select * from TEST_TABLE')  # doctest: +SKIP
        >>> otp.run(data)  # doctest: +SKIP
                Time   A     B        C                       D
        0 2003-12-04  A1  1975  8.12345 2022-01-01 12:13:14.111
        1 2003-12-04  A2  1971  7.98765 2022-01-02 22:23:24.222

        Substitute start time placeholder with ``start_expr`` parameter:

        >>> data = otp.ODBC(
        ...     dsn='testdb_dsn',
        ...     sql='select * from TEST_TIMESTAMP where TIMESTAMP >= "<_START_TIME>"',
        ...     start_expr=(otp.meta_fields['START_TIME'] + otp.Day(1)).dt.strftime('%Y-%m-%d %H:%M:%S.%q')
        ... )  # doctest: +SKIP
        >>> otp.run(data, start=otp.dt(2022, 1, 1), end=otp.dt(2022, 1, 3))  # doctest: +SKIP
                             Time  A
        0 2022-01-02 22:23:24.222  2

        Use parameter ``allow_unordered_ticks`` if needed:

        >>> data = otp.ODBC(dsn='testdb_dsn',
        ...                 sql='select * from TEST_UNORDERED',
        ...                 allow_unordered_ticks=True)  # doctest: +SKIP
        >>> otp.run(data, start=otp.dt(2022, 1, 1), end=otp.dt(2022, 1, 3))  # doctest: +SKIP
                             Time  A
        0 2022-01-02 22:23:24.222  2
        1 2022-01-01 12:13:14.111  1
        """
        if self._try_default_constructor(schema=schema, **kwargs):
            return

        if not is_odbc_query_supported():
            raise RuntimeError("ODBC source is not supported in this version of OneTick, "
                               "it is available starting from release 1.24 (or development build 20231108)")

        if not dsn and not connection_string:
            raise ValueError("One of the parameters 'dsn' or 'connection_string' must be specified")

        if authentication_type not in {'system', 'credentials'}:
            raise ValueError(f"Unknown value for parameter 'authentication_type': {authentication_type}")

        if connection_string and any([dsn, user, password, authentication_type == 'system']):
            raise ValueError("Parameter 'connection_string' is used instead of parameters"
                             " 'dsn', 'user', 'password' and 'authentication_type'")

        if not sql:
            raise ValueError("Parameter 'sql' must be set")

        if not isinstance(numeric_scale, int) or not 0 <= numeric_scale <= 34:
            raise ValueError("Unsupported value for parameter 'numeric_scale'")

        self._dsn = dsn
        self._connection_string = connection_string
        self._authentication_type = authentication_type
        self._user = user
        self._password = password
        self._sql = sql
        self._symbology = symbology
        self._allow_unordered_ticks = allow_unordered_ticks
        self._apply_symbol_name_history = apply_symbol_name_history
        self._numeric_scale = numeric_scale
        self._preserve_unicode_fields = preserve_unicode_fields
        self._tz = tz
        self._start_expr = start_expr
        self._end_expr = end_expr
        self._db = db
        self._tick_type = tick_type
        self._symbols = symbols
        self._presort = presort

        # if we have a list of symbols, they will be merged in self.base_ep()
        if isinstance(symbols, list):
            symbols = None

        super().__init__(
            _symbols=symbols, _start=start, _end=end, _base_ep_func=lambda: self.base_ep(schema=schema, **kwargs),
            schema=schema, **kwargs,
        )

    def base_ep(self, schema=None, **kwargs):

        src = Source(
            otq.Omd_odbcQuery(
                dsn=self._dsn or '',
                connection_string=self._connection_string or '',
                authentication_type='System' if self._authentication_type == 'system' else 'Username/Password',
                user=self._user or '',
                password=self._password or '',
                sql=self._sql,
                symbology=self._symbology or '',
                allow_unordered_ticks=self._allow_unordered_ticks,
                apply_symbol_name_history=self._apply_symbol_name_history,
                numeric_scale=self._numeric_scale,
                preserve_unicode_fields=','.join(self._preserve_unicode_fields or []),
                tz=self._tz or configuration.config.get('tz') or '',
                start_time_expr=str(self._start_expr) if self._start_expr is not None else '',
                end_time_expr=str(self._end_expr) if self._end_expr is not None else '',
            ),
            schema=schema,
            **kwargs,
        )

        db = self._db

        if isinstance(self._symbols, list):
            if self._presort:
                src.sink(otq.Presort().symbols(self._symbols))
                src.sink(otq.Merge(identify_input_ts=False))
            else:
                src.sink(otq.Merge(identify_input_ts=False).symbols(self._symbols))

        update_node_tick_type(src, self._tick_type, db)

        return src



from typing import TYPE_CHECKING, Callable, List, Optional, Tuple, Union

from onetick import py as otp
from onetick.py import configuration
from onetick.py.otq import otq

from .misc import inplace_operation

if TYPE_CHECKING:
    from onetick.py.core.source import Source


@inplace_operation
def throw(
    self: 'Source', message='', where=1, scope='query', error_code=1, throw_before_query_execution=False, inplace=False
) -> Optional['Source']:
    """
    Propagates error or warning or throws an exception (depending on ``scope`` parameter)
    when condition provided by ``where`` parameter evaluates to True.

    Throwing an exception will abort query execution,
    propagating error will stop tick propagation and
    propagating warning will not change the data processing.

    Parameters
    ----------
    message: str or Operation
        Message that will be thrown in exception or returned in error message.
    where: Operation
        Logical expression that specifies condition when exception or error will be thrown.
    scope: 'query' or 'symbol'
        'query' will throw an exception and 'symbol' will propagate an error or warning
        depending on ``error_code`` parameter.
    error_code: int
        When ``scope='symbol'``, values from interval ``[1, 500]`` indicate warnings
        and values from interval ``[1500, 2000]`` indicate errors.
        Note that tick propagation will not stop when warning is raised,
        and will stop when error is raised.

        This parameter is not used when ``scope='query'``.
    throw_before_query_execution: bool
        If set to ``True``, the exception will be thrown before the execution of the query.
        This option is intended for supplying placeholder queries that must always throw.
    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing.
        Otherwise method returns a new modified object.

    See also
    --------
    **THROW** OneTick event processor

    Returns
    -------
    :class:`Source` or ``None``

    Examples
    --------

    By default, this method will throw an exception on the first tick with empty message:

    >>> t = otp.Tick(A=1)
    >>> t = t.throw()
    >>> otp.run(t)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    Exception: ...: In THROW: ...
        ...

    You can specify exception message and condition (note that exception now is raised on second tick):

    >>> t = otp.Ticks(A=[1, 2])
    >>> t = t.throw(message='A is ' + t['A'].apply(str), where=(t['A']==2))
    >>> otp.run(t)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    Exception: ...: In THROW: A is 2. ...
        ...

    Note that exception will not be thrown if there are no ticks:

    >>> t = otp.Empty()
    >>> t = t.throw()
    >>> otp.run(t)
    Empty DataFrame
    Columns: []
    Index: []

    If you need exception to be thrown always, you can use ``throw_before_query_execution`` parameter:

    >>> t = otp.Empty()
    >>> t = t.throw(throw_before_query_execution=True)
    >>> otp.run(t)  # doctest: +ELLIPSIS
    Traceback (most recent call last):
        ...
    Exception: ...: In THROW: ...

    You can throw OneTick errors and warnings instead of exceptions.
    Raising error codes from 1 to 500 indicates warnings.
    Raising error codes from 1500 to 2000 indicates errors.
    Tick propagation stops only when error is raised.

    >>> t = otp.Ticks(A=[1, 2, 3, 4])
    >>> t = t.throw(message='warning A=1', scope='symbol', error_code=2, where=(t['A']==1))
    >>> t = t.throw(message='error A=3', scope='symbol', error_code=1502, where=(t['A']==3))
    >>> otp.run(t)  # doctest: +SKIP
    UserWarning: Symbol error: [2] warning A=1
    UserWarning: Symbol error: [1502] error A=3
                         Time  A
    0 2003-12-01 00:00:00.000  1
    1 2003-12-01 00:00:00.001  2

    Right now the only supported interface to get errors and warnings
    is via ``map`` output structure and its methods:

    >>> otp.run(t, symbols='AAPL', output_structure='map').output('AAPL').error
    [(2, 'warning A=1'), (1502, 'error A=3')]
    """
    if isinstance(message, str):
        treat_message_as_per_tick_expr = False
    else:
        treat_message_as_per_tick_expr = True
    if scope not in {'query', 'symbol'}:
        raise ValueError("Parameter 'scope' can only be set to 'query' or 'symbol'")
    if not (1 <= error_code <= 500 or 1500 <= error_code <= 2000):
        raise ValueError(
            "Parameter 'error_code' can only take values "
            "from interval [1, 500] for warnings and from interval [1500, 2000] for errors"
        )
    self.sink(
        otq.Throw(
            message=str(message),
            treat_message_as_per_tick_expr=treat_message_as_per_tick_expr,
            where=str(where),
            scope=scope.upper(),
            error_code=error_code,
            throw_before_query_execution=throw_before_query_execution,
        )
    )
    return self


@inplace_operation
def logf(
    self: 'Source', message: str, severity: str, *args, where: Optional['otp.Operation'] = None, inplace: bool = False
) -> Optional['Source']:
    """
    Call built-in OneTick ``LOGF`` function.

    Parameters
    ----------
    message: str
        Log message/format string. The underlying formatting engine is the Boost Format Library:
        https://www.boost.org/doc/libs/1_53_0/libs/format/doc/format.html

    severity: str
        Severity of message. Supported values: ``ERROR``, ``WARNING`` and ``INFO``.

    where: Operation
        A condition that allows to filter ticks to call ``LOGF`` on. ``None`` for no filtering.

    inplace: bool
        The flag controls whether operation should be applied inplace or not.
        If ``inplace=True``, then it returns nothing. Otherwise method returns a new modified
        object.

    args: list
        Parameters for format string (optional).

    Returns
    -------
    :class:`Source`

    Examples
    --------
    >>> data = otp.Ticks(PRICE=[97, 99, 103])
    >>> data = data.logf("PRICE (%1%) is higher than 100", "WARNING", data["PRICE"], where=(data["PRICE"] > 100))
    """

    where_in_tick = where if where is not None else True

    def per_tick_script(tick):
        if where_in_tick:
            otp.logf(message, severity, *args)

    return self.script(per_tick_script, inplace=inplace)


def dump(
    self: 'Source',
    label: Optional[str] = None,
    where: Optional['otp.Operation'] = None,
    columns: Union[str, Tuple[str], List[str], None] = None,
    callback: Optional[Callable] = None,
):
    """
    Dumps the ``columns`` from ticks into std::out in runtime if they fit the ``where`` condition.
    Every dump has a corresponding header that always includes the TIMESTAMP field. Other fields
    could be configured using the ``columns`` parameter. A header could be augmented with a ``label`` parameter;
    this label is an addition column that helps to distinguish ticks
    from multiple dumps with the same schema, because ticks from different dumps could be mixed.
    It might happen because of the OneTick multithreading, and there is the operating system
    buffer between the OneTick and the actual output.

    This method is helpful for debugging.

    Parameters
    ----------
    label: str
        A label for a dump. It adds a special column **_OUT_LABEL_** for all ticks and set to the
        specified value. It helps to distinguish ticks from multiple dumps, because actual
        output could contain mixed ticks due the concurrency. ``None`` means no label.
    where: Operation
        A condition that allows to filter ticks to dump. ``None`` means no filtering.
    columns: str, tupel or list
        List of columns that should be in the output. ``None`` means dump all columns.
    callback : callable
        Callable, which preprocess source before printing.

    See also
    --------
    **WRITE_TEXT** OneTick event processor

    Examples
    --------
    >>> # OTdirective: skip-snippet:;
    >>> data.dump(label='Debug point', where=data['PRICE'] > 99.3, columns=['PRICE', 'QTY'])    # doctest: +SKIP
    >>> data.dump(columns="X", callback=lambda x: x.first(), label="first")     # doctest: +SKIP
    """

    self_c = self.copy()
    if callback:
        self_c = callback(self_c)
    if where is not None:  # can't be simplified because the _Operation overrides __bool__
        self_c, _ = self_c[(where)]
    if columns:
        self_c = self_c[columns if isinstance(columns, (list, tuple)) else [columns]]
    if label:
        self_c['_OUT_LABEL_'] = label

    self_c.write_text(
        formats_of_fields={'TIMESTAMP': f'%|{configuration.config.tz}|%d-%m-%Y %H:%M:%S.%J'},
        prepend_timestamp=True,
        prepend_symbol_name=False,
        propagate_ticks=True,
        inplace=True,
    )

    # print <no data> in case there are 0 ticks
    if hasattr(otq, 'InsertAtEnd'):
        ep_kwargs = {}
        if {'propagate_ticks', 'insert_if_input_is_empty', 'fields_of_added_tick'}.issubset(
            otq.InsertAtEnd.Parameters.list_parameters()
        ):
            # if supported, fix PY-1433
            ep_kwargs = dict(
                propagate_ticks=True,
                insert_if_input_is_empty=True,
                fields_of_added_tick='AT_END=1'
            )
        self_c.sink(otq.InsertAtEnd(delimiter_name='AT_END', **ep_kwargs))
        self_c.schema['AT_END'] = int
        self_c.state_vars['COUNT'] = otp.state.var(0, scope='branch')
        self_c.state_vars['COUNT'] += 1
        self_c, _ = self_c[(self_c['AT_END'] == 1) & (self_c.state_vars['COUNT'] == 1)]
        self_c['NO_DATA'] = '<no data>'
        self_c = self_c[['NO_DATA']]
        self_c.write_text(output_headers=False, prepend_timestamp=False, prepend_symbol_name=False,
                          propagate_ticks=False, inplace=True)

    # Do not propagate ticks then, because we want just to write them into
    # the std::out. We have to do that, because otherwise these ticks would
    # go to a query output, that would mess real output.
    self_c, _ = self_c[self_c['Time'] != self_c['Time']]  # NOSONAR

    # We have to merge the branch back to the main branch even that these
    # branch does not generate ticks, because we do not introduce one more
    # output point, because the OneTick would add it to the final output
    # datastructure.
    self.sink(otq.Merge(identify_input_ts=False))

    self.source(self_c.node().copy_graph())  # NOSONAR
    self.node().add_rules(self_c.node().copy_rules())
    self._merge_tmp_otq(self_c)



import warnings

from onetick.py import types as ott
from onetick.py.core._internal._state_objects import (
    _StateColumn, TickList, TickSet, TickSetUnordered, TickDeque,
    _TickListTick, _TickSetTick, _TickDequeTick, _DynamicTick,
)


AVAILABLE_SCOPES = (
    "BRANCH",  # A branch-scope state variable is visible to all event processors of the branch its declarator EP
    # belongs to (a branch is a maximal long chain of EPs, each node of which has at most 1 _source and at most 1 sink).
    "ALL_INPUTS",  # An all-inputs-scope state variable is visible to all event processors of the input subgraph of its
    # declarator EP (input subgraph of a particular node is the subset of nodes directly or indirectly feeding it).
    "ALL_OUTPUTS",  # An all-outputs-scope state variable is visible to all event processors of the output subgraph of
    # its declarator EP (output subgraph of a particular node is the subset of nodes directly or indirectly fed by it).
    "QUERY",  # A query scope state variable is visible to all event processors of the graph.
    "CROSS_SYMBOL",  # A cross-symbol scope state variable is visible to all event processors across all unbound
    # symbols. Cross-symbol scope state variables cannot be modified after initialization.
)


def var(default_value, scope="query"):
    """
    Defines a state variable. Supports int, float and string values.

    Parameters
    ----------
    default_value: any
        Default value of the state variable
    scope: str
        Scope for the state variable. Possible values are: query, branch, cross_symbol.
        Default: query

    Returns
    -------
    a state variable that should be assigned to a _source

    Examples
    --------
    >>> data = otp.Ticks(dict(X=[0, 1, 2]))
    >>> data.state_vars['SUM'] = otp.state.var(0)
    >>> data.state_vars['SUM'] += data['X']
    >>> data['SUM'] = data.state_vars['SUM']
    >>> otp.run(data)[['X', 'SUM']]
       X  SUM
    0  0    0
    1  1    1
    2  2    3
    """
    scope = _validate_and_preprocess_scope(scope)

    dtype = ott.get_object_type(default_value)

    if dtype is ott.msectime:
        raise TypeError("State variables do not support msectime type")
    # elif dtype is ott.nsectime:
    #     raise TypeError("State variables do not support nsectime type")
    elif ott.is_time_type(dtype):
        dtype = ott.nsectime

    # pylint: disable-next=unidiomatic-typecheck
    if type(default_value) is str:
        if len(default_value) > ott.string.DEFAULT_LENGTH:
            dtype = ott.string[len(default_value)]

    res = _StateColumn("__STATE_COLUMN", dtype, obj_ref=None, default_value=default_value, scope=scope)

    return res


def tick_list(default_value=None, scope='query', schema=None) -> TickList:
    """
    Defines a state tick list.

    Parameters
    ----------
    default_value: :class:`otp.Source <onetick.py.Source>`, :py:func:`eval query <onetick.py.eval>`
        Evaluated query to initialize tick list from.
    scope: str
        Scope for the state variable.
        Possible values are: query, branch, cross_symbol, all_inputs, all_outputs
    schema: dict, list
        Desired schema for the created tick list. If not passed, schema will be inherited from ``default_value``.
        if ``default_value`` is not passed as well, schema will contain fields of the main Source object.

        If schema is passed as a list, it will select only these fields from the schema of ``default_value``
        or main :class:`Source <onetick.py.Source>` object.

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data.state_vars['LIST'] = otp.state.tick_list(otp.Ticks(B=[1, 2, 3]))
    >>> data = data.state_vars['LIST'].dump()
    >>> otp.run(data)[['B']]
       B
    0  1
    1  2
    2  3
    """
    scope = _validate_and_preprocess_scope(scope)
    return TickList('', obj_ref=None, default_value=default_value, scope=scope, schema=schema)


def tick_set(insertion_policy, key_fields, default_value=None, scope='query', schema=None) -> TickSet:
    """
    Defines a state tick set.

    Parameters
    ----------
    insertion_policy: 'oldest' or 'latest'
        'oldest' specifies not to overwrite ticks with the same keys.
        'latest' makes the last inserted tick overwrite the one with the same keys (if existing).
    key_fields: str, list of str
        The values of the specified fields will be used as keys.
    default_value: :class:`otp.Source <onetick.py.Source>`, :py:func:`eval query <onetick.py.eval>`
        Evaluated query to initialize tick set from.
    scope: str
        Scope for the state variable.
        Possible values are: query, branch, cross_symbol, all_inputs, all_outputs
    schema: dict, list
        Desired schema for the created tick set. If not passed, schema will be inherited from ``default_value``.
        if ``default_value`` is not passed as well, schema will contain fields of the main Source object.

        If schema is passed as a list, it will select only these fields from the schema of ``default_value``
        or main :class:`Source <onetick.py.Source>` object.

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data.state_vars['SET'] = otp.state.tick_set('oldest', 'B', otp.Ticks(B=[1, 1, 2, 2, 3, 3]))
    >>> data = data.state_vars['SET'].dump()
    >>> otp.run(data)[['B']]
       B
    0  1
    1  2
    2  3
    """
    scope = _validate_and_preprocess_scope(scope)
    return TickSet('',
                   insertion_policy=insertion_policy, key_fields=key_fields,
                   obj_ref=None, default_value=default_value, scope=scope,
                   schema=schema)


def tick_set_unordered(insertion_policy,
                       key_fields,
                       default_value=None,
                       max_distinct_keys=-1,
                       scope='query',
                       schema=None) -> TickSetUnordered:
    """
    Defines unordered tick set.

    Parameters
    ----------
    insertion_policy: 'oldest' or 'latest'
        'oldest' specifies not to overwrite ticks with the same keys.
        'latest' makes the last inserted tick overwrite the one with the same keys (if existing).
    key_fields: str, list of str
        The values of the specified fields will be used as keys.
    max_distinct_keys: int
        Expected size of the unordered tick set. It will be used to allocate memory to hold ticks in the tick set.
        It should have the same order as the expected amounts of ticks held in the set. Default value of -1 indicates
        that expected amount of ticks is not known in advance.
        If this value is set correctly, the performance of unordered tick set is expected to be better
        than that of normal tick set.

        .. warning::
            If ``max_distinct_keys`` value is set too low, the performance of unordered tick set
            may be considerably worse than that of normal tick set.
            In particular, **default value of -1 will lead to bad performance and should be avoided**

    default_value: :class:`otp.Source <onetick.py.Source>`, :py:func:`eval query <onetick.py.eval>`
        Evaluated query to initialize unordered tick set from.
    scope: str,
        Scope for the state variable.
        Possible values are: query, branch, cross_symbol, all_inputs, all_outputs
    schema: dict, list
        Desired schema for the created unordered tick set. If not passed,
        schema will be inherited from ``default_value``.
        if ``default_value`` is not passed as well, schema will contain fields of the main Source object.

        If schema is passed as a list, it will select only these fields from the schema of ``default_value``
        or main :class:`Source <onetick.py.Source>` object.

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data.state_vars['SET'] = otp.state.tick_set_unordered('oldest', 'B',
    ...                                                       otp.Ticks(B=[1, 1, 2, 2, 3, 3]),
    ...                                                       max_distinct_keys=5)
    >>> data = data.state_vars['SET'].dump()
    >>> otp.run(data)[['B']]
       B
    0  1
    1  2
    2  3
    """
    scope = _validate_and_preprocess_scope(scope)
    if max_distinct_keys == -1:
        warnings.warn('Unordered tick set with max_distinct_keys == -1 is expected to be inefficient. '
                      'Please consider using regular tick set instead.', stacklevel=2)
    return TickSetUnordered(
        '',
        insertion_policy=insertion_policy, max_distinct_keys=max_distinct_keys, key_fields=key_fields,
        obj_ref=None, default_value=default_value, scope=scope, schema=schema
    )


def tick_deque(default_value=None, scope='query', schema=None) -> TickDeque:
    """
    Defines a state tick deque.

    Parameters
    ----------
    default_value: :class:`otp.Source <onetick.py.Source>`, :py:func:`eval query <onetick.py.eval>`
        Evaluated query to initialize tick deque from.
    scope: str
        Scope for the state variable.
        Possible values are: query, branch, cross_symbol, all_inputs, all_outputs
    schema: dict, list
        Desired schema for the created tick deque. If not passed, schema will be inherited from ``default_value``.
        if ``default_value`` is not passed as well, schema will contain fields of the main Source object.

        If schema is passed as a list, it will select only these fields from the schema of ``default_value``
        or main :class:`Source <onetick.py.Source>` object.

    Examples
    --------
    >>> data = otp.Tick(A=1)
    >>> data.state_vars['DEQUE'] = otp.state.tick_deque(otp.Ticks(B=[1, 2, 3]))
    >>> data = data.state_vars['DEQUE'].dump()
    >>> otp.run(data)[['B']]
       B
    0  1
    1  2
    2  3
    """
    scope = _validate_and_preprocess_scope(scope)
    return TickDeque('', obj_ref=None, default_value=default_value, scope=scope, schema=schema)


def _validate_and_preprocess_scope(scope):
    if isinstance(scope, str):
        scope = scope.upper()
        if scope not in AVAILABLE_SCOPES:
            raise ValueError(f"unknown scope {scope}, please use one of {AVAILABLE_SCOPES}")
    else:
        raise ValueError(f"scope should be one of the following strings: {AVAILABLE_SCOPES}")
    return scope



import argparse
from typing import Any, Dict, Optional

import onetick.py as otp


def parse_args():
    parser = argparse.ArgumentParser(description='Render queries from otq file')

    parser.add_argument('path', help='Path to otq file')
    parser.add_argument(
        '--image-path',
        help='Path to output image file. Default: query.svg in current working directory',
    )
    parser.add_argument('--output-format', help='Output format of output image')
    parser.add_argument(
        '--load-external-otqs', help='Load query dependencies from external otq files',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--view', help='Show generated image after render', action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--y-limit', help='Limit for maximum number of lines of some EP parameters strings', type=int,
    )
    parser.add_argument(
        '--x-limit', help='Limit for maximum number of characters for each line in text of some EP parameters strings',
        type=int,
    )
    parser.add_argument(
        '--parse-eval-from-params', help='Enable parsing and printing `eval` sub-queries from EP parameters',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--render-debug-info', help='Render additional debug information',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--debug', help='Allow to print stdout or stderr from `Graphviz` render',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--graphviz-compat-mode',
        help='Change internal parameters of result graph for better compatibility with old `Graphviz` versions',
        action=argparse.BooleanOptionalAction,
    )
    parser.add_argument(
        '--font-family', help='Font family',
        type=str,
    )
    parser.add_argument(
        '--font-size', help='Font size',
        type=int,
    )

    args = parser.parse_args()
    return args


def render_otq(
    path: str,
    image_path: Optional[str] = None,
    y_limit: Optional[int] = None,
    x_limit: Optional[int] = None,
    **kwargs,
):
    if image_path is None:
        if kwargs.get('output_format'):
            raise ValueError('`image-path` should be specified in order to use parameter `output-format`')

        image_path = 'query.svg'

    call_kwargs: Dict[str, Any] = {'path': path, 'image_path': image_path}
    if x_limit is not None and y_limit is not None:
        call_kwargs['line_limit'] = (y_limit, x_limit)
    elif x_limit is not None or y_limit is not None:
        raise ValueError('Both `y_limit` and `x_limit` should be set')

    for param_name in [
        'output_format', 'load_external_otqs', 'view', 'parse_eval_from_params', 'render_debug_info',
        'debug', 'graphviz_compat_mode', 'font_family', 'font_size',
    ]:
        if kwargs.get(param_name) is not None:
            call_kwargs[param_name] = kwargs[param_name]

    output_path = otp.utils.render_otq(**call_kwargs)
    print(f'Rendered graph saved in {output_path}')


def main():
    args = parse_args()
    render_otq(**vars(args))


if __name__ == '__main__':
    main()
